__all__ = ['ncf2cloud_rain']

import numpy as np
from PseudoNetCDF._getwriter import registerwriter


def ncf2cloud_rain(ncffile, outpath, tflag='TFLAG'):
    outfile = open(outpath, 'wb')
    varkeys = [vk for vk in ['CLOUD', 'PRECIP', 'RAIN', 'SNOW',
                             'GRAUPEL', 'COD']
               if vk in ncffile.variables.keys()]
    buf = np.array(len(ncffile.FILEDESC) + 12, dtype='>i').tostring()
    outfile.write(buf)
    outfile.write(np.array(ncffile.FILEDESC, dtype='c'))
    nxcl = len(ncffile.dimensions['COL'])
    nycl = len(ncffile.dimensions['ROW'])
    nzcl = len(ncffile.dimensions['LAY'])
    outfile.write(np.array([nxcl, nycl, nzcl], dtype='>i'))
    outfile.write(buf)

    for di, (d, t) in enumerate(ncffile.variables[tflag][:, 0, :]):
        t = np.array(t.astype('>f') / 100, ndmin=1).astype('>f')
        d = np.array(d, ndmin=1).astype('>i')
        d = (d % (d // 100000 * 100000)).astype('>i')
        buf = np.array(8, dtype='>i').tostring()
        outfile.write(buf + t.tostring() + d.tostring() + buf)
        for zi in range(nzcl):
            for varkey in varkeys:
                vals = ncffile.variables[varkey][di, zi].astype('>f')
                buf = np.array((vals.size) * 4, ndmin=1).astype('>i')
                buf = buf.tostring()
                outfile.write(buf)
                vals.tofile(outfile)
                outfile.write(buf)

    outfile.flush()
    return outfile


registerwriter('camxfiles.cloud_rain', ncf2cloud_rain)
registerwriter('cloud_rain', ncf2cloud_rain)



__all__ = ['ipr']
__doc__ = """
.. _Memmap
:mod:`Memmap` -- ipr Memmap interface
============================================

.. module:: Memmap
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` memory map for CAMx
              ipr files.  See PseudoNetCDF.sci_var.PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""
# Distribution packages
import unittest
import struct
import functools
from warnings import warn

# Site-Packages
from numpy import zeros, dtype, memmap, char

# This Package modules
from PseudoNetCDF.conventions.ioapi import add_cf_from_ioapi
from PseudoNetCDF.camxfiles.timetuple import timeadd, timerange
from PseudoNetCDF.camxfiles.util import cartesian
from PseudoNetCDF.camxfiles.units import get_uamiv_units
from PseudoNetCDF.camxfiles.FortranFileUtil import OpenRecordFile
from PseudoNetCDF.sci_var import PseudoNetCDFFile, PseudoNetCDFVariable
from PseudoNetCDF.sci_var import PseudoNetCDFVariables
from PseudoNetCDF.ArrayTransforms import ConvertCAMxTime

# for use in identifying uncaught nan
listnan = struct.unpack('>f', b'\xff\xc0\x00\x00')[0]
checkarray = zeros((1,), 'f')
checkarray[0] = listnan
array_nan = checkarray[0]


class ipr(PseudoNetCDFFile):
    """
    ipr provides a PseudoNetCDF interface for CAMx
    ipr files.  Where possible, the inteface follows
    IOAPI conventions (see www.baronams.com).

    ex:
        >>> ipr_path = 'camx_ipr.bin'
        >>> iprfile = ipr(ipr_path)
        >>> iprfile.variables.keys()
        ['TFLAG', 'SPAD_O3', 'DATE_O3', 'TIME_O3', 'SPC_O3',
         'PAGRID_O3', 'NEST_O3', 'I_O3', 'J_O3', 'K_O3',
         'INIT_O3', 'CHEM_O3', 'EMIS_O3', 'PTEMIS_O3',
         'PIG_O3', 'WADV_O3', 'EADV_O3', 'SADV_O3', 'NADV_O3',
         'BADV_O3', 'TADV_O3', 'DIL_O3', 'WDIF_O3', 'EDIF_O3',
         'SDIF_O3', 'NDIF_O3', 'BDIF_O3', 'TDIF_O3', 'DDEP_O3',
         'WDEP_O3', 'INORGACHEM_O3', 'ORGACHEM_O3', 'AQACHEM_O3',
         'FCONC_O3', 'UCNV_O3', 'AVOL_O3', 'EPAD_O3']
        >>> v = iprfile.variables['CHEM_O3']
        >>> tflag = iprfile.variables['TFLAG']
        >>> tflag.dimensions
        ('TSTEP', 'VAR', 'DATE-TIME')
        >>> tflag[0,0,:]
        array([2005185,       0])
        >>> tflag[-1,0,:]
        array([2005185,  240000])
        >>> v.dimensions
        ('TSTEP', 'LAY', 'ROW', 'COL')
        >>> v.shape
        (25, 28, 65, 83)
        >>> iprfile.dimensions
        {'TSTEP': 25, 'LAY': 28, 'ROW': 65, 'COL': 83}
    """

    id_fmt = "if10s5i"
    dt_fmt = "if"
    data_fmt = "f"

    def __init__(self, rf, multi=False, **props):
        """
        Initialization included reading the header and learning
        about the format.

        see __readheader and __gettimestep() for more info

        Keywords (i.e., props) for projection: P_ALP, P_BET, P_GAM, XCENT,
                                            YCENT, XORIG, YORIG, XCELL, YCELL
        """
        self.__rffile = OpenRecordFile(rf)
        self.__readheader()
        self.__ipr_record_type = {
            24: dtype(
                dict(
                    names=['SPAD', 'DATE', 'TIME', 'SPC', 'PAGRID', 'NEST',
                           'I', 'J', 'K', 'INIT', 'CHEM', 'EMIS', 'PTEMIS',
                           'PIG', 'WADV', 'EADV', 'SADV', 'NADV', 'BADV',
                           'TADV', 'DIL', 'WDIF', 'EDIF', 'SDIF', 'NDIF',
                           'BDIF', 'TDIF', 'DDEP', 'WDEP', 'AERCHEM', 'FCONC',
                           'UCNV', 'AVOL', 'EPAD'],
                    formats=['>i', '>i', '>f', '>S10', '>i', '>i', '>i', '>i',
                             '>i', '>f', '>f', '>f', '>f', '>f', '>f', '>f',
                             '>f', '>f', '>f', '>f', '>f', '>f', '>f', '>f',
                             '>f', '>f', '>f', '>f', '>f', '>f', '>f', '>f',
                             '>f', '>i'])),
            26: dtype(
                dict(
                    names=['SPAD', 'DATE', 'TIME', 'SPC', 'PAGRID', 'NEST',
                           'I', 'J', 'K', 'INIT', 'CHEM', 'EMIS', 'PTEMIS',
                           'PIG', 'WADV', 'EADV', 'SADV', 'NADV', 'BADV',
                           'TADV', 'DIL', 'WDIF', 'EDIF', 'SDIF', 'NDIF',
                           'BDIF', 'TDIF', 'DDEP', 'WDEP', 'INORGACHEM',
                           'ORGACHEM', 'AQACHEM', 'FCONC', 'UCNV', 'AVOL',
                           'EPAD'],
                    formats=['>i', '>i', '>f', '>S10', '>i', '>i', '>i', '>i',
                             '>i', '>f', '>f', '>f', '>f', '>f', '>f', '>f',
                             '>f', '>f', '>f', '>f', '>f', '>f', '>f', '>f',
                             '>f', '>f', '>f', '>f', '>f', '>f', '>f', '>f',
                             '>f', '>f', '>f', '>i']))
        }[len(self.prcnames)]

        aeroprocs = {24: ['AERCHEM'],
                     26: ['INORGACHEM', 'ORGACHEM',
                          'AQACHEM']}[len(self.prcnames)]
        prcs = (['SPAD', 'DATE', 'TIME', 'PAGRID', 'NEST', 'I', 'J', 'K',
                 'INIT', 'CHEM', 'EMIS', 'PTEMIS', 'PIG', 'WADV', 'EADV',
                 'SADV', 'NADV', 'BADV', 'TADV', 'DIL', 'WDIF', 'EDIF', 'SDIF',
                 'NDIF', 'BDIF', 'TDIF', 'DDEP', 'WDEP'] +
                aeroprocs + ['FCONC', 'UCNV', 'AVOL', 'EPAD'])
        varkeys = ['_'.join(i) for i in cartesian(prcs, self.spcnames)]
        varkeys += ['SPAD', 'DATE', 'TIME',
                    'PAGRID', 'NEST', 'I', 'J', 'K', 'TFLAG']
        self.groups = {}
        NSTEPS = len([i_ for i_ in self.timerange()])
        NVARS = len(varkeys)
        self.createDimension('VAR', NVARS)
        self.createDimension('DATE-TIME', 2)
        self.createDimension('TSTEP', NSTEPS)
        padatatype = []
        pavarkeys = []
        for di, domain in enumerate(self.padomains):
            dk = 'PA%02d' % di
            prefix = dk + '_'
            if len(self.padomains) == 1:
                grp = self.groups[dk] = self
            else:
                grp = self.groups[dk] = PseudoNetCDFFile()
            pavarkeys.extend([prefix + k for k in varkeys])
            grp.createDimension('VAR', NVARS)
            grp.createDimension('DATE-TIME', 2)
            grp.createDimension('TSTEP', NSTEPS)
            grp.createDimension('COL', domain['iend'] - domain['istart'] + 1)
            grp.createDimension('ROW', domain['jend'] - domain['jstart'] + 1)
            grp.createDimension('LAY', domain['tlay'] - domain['blay'] + 1)
            padatatype.append((dk, self.__ipr_record_type,
                              (len(grp.dimensions['ROW']),
                               len(grp.dimensions['COL']),
                               len(grp.dimensions['LAY']))))
            grp.varget = functools.partial(self.__variables, dk)
            # grp.varget = eval("lambda k: self._ipr__variables('%s', k)""" %
            #                   dk, dict(self=self), locals())
            if len(self.padomains) == 1:
                self.variables = PseudoNetCDFVariables(grp.varget, varkeys)
            else:
                grp.variables = PseudoNetCDFVariables(grp.varget, varkeys)
        nspc = len(self.spcnames)
        self.__memmaps = memmap(self.__rffile.infile.name, dtype(padatatype),
                                'r',
                                self.data_start_byte).reshape(NSTEPS, nspc)
        for k, v in props.items():
            setattr(self, k, v)
        try:
            add_cf_from_ioapi(self)
        except Exception:
            pass

    def __del__(self):
        try:
            del self.__memmaps
        except Exception:
            pass

    def __decorator(self, name, pncfv):
        spc = name.split('_')[-1]
        prc = name.split('_')[0]
        # IPR units are consistent with 'IPR'
        if prc == 'UCNV':
            units = 'm**3/mol'
        elif prc == 'AVOL':
            units = 'm**3'
        else:
            units = get_uamiv_units('IPR', spc)

        def decor(k):
            return dict(units=units, var_desc=k.ljust(16),
                        long_name=k.ljust(16))
        for k, v in decor(name).items():
            setattr(pncfv, k, v)
        return pncfv

    def __variables(self, pk, proc_spc):
        if proc_spc in self.__ipr_record_type.names:
            proc = proc_spc
            proc_spc = proc_spc + '_' + self.spcnames[0]
            tmpvals = self.__memmaps[pk][:, 0, :, :, :][proc]
            tmpvals = tmpvals.swapaxes(1, 3).swapaxes(2, 3)
            return PseudoNetCDFVariable(self, proc_spc, 'f',
                                        ('TSTEP', 'LAY', 'ROW', 'COL'),
                                        values=tmpvals)
        if proc_spc == 'TFLAG':
            thisdate = self.__memmaps[pk][:, 0, :, :, :]['DATE'].swapaxes(
                1, 3).swapaxes(2, 3)[..., 0, 0, 0]
            thistime = self.__memmaps[pk][:, 0, :, :, :]['TIME'].swapaxes(
                1, 3).swapaxes(2, 3)[..., 0, 0, 0]
            nvar = len(self.groups[pk].dimensions['VAR'])
            return ConvertCAMxTime(thisdate, thistime, nvar)
        for k in self.__ipr_record_type.names:
            proc = proc_spc[:len(k)]
            spc = proc_spc[len(k) + 1:]
            if proc == k and spc in self.spcnames:
                spc = self.spcnames.index(spc)
                dvals = self.__memmaps[pk][:, spc][proc].swapaxes(
                    1, 3).swapaxes(2, 3)
                tmpvar = PseudoNetCDFVariable(self, proc_spc, 'f',
                                              ('TSTEP', 'LAY', 'ROW', 'COL'),
                                              values=dvals)
                return self.__decorator(proc_spc, tmpvar)
        raise KeyError("Bad!")

    def __readheader(self):
        """
        __readheader reads the header section of the ipr file
        it initializes each header field (see CAMx Users Manual for a list)
        as properties of the ipr class
        """

        self.runmessage = self.__rffile.read("80s")
        self.start_date, self.start_time, self.end_date, self.end_time = \
            self.__rffile.read("ifif")

        self.grids = []
        for grid in range(self.__rffile.read("i")[-1]):
            self.grids.append(
                dict(
                    zip(
                        ['orgx', 'orgy', 'ncol', 'nrow', 'xsize', 'ysize'],
                        self.__rffile.read("iiiiii")
                    )
                )
            )

        self.spcnames = []
        for spc in range(self.__rffile.read("i")[-1]):
            self.spcnames.append(char.decode(
                self.__rffile.read("10s")[-1]).strip())

        self.nspec = len(self.spcnames)
        self.padomains = []

        for padomain in range(self.__rffile.read("i")[-1]):
            self.padomains.append(
                dict(
                    zip(
                        ['grid', 'istart', 'iend', 'jstart', 'jend', 'blay',
                         'tlay'],
                        self.__rffile.read("iiiiiii")
                    )
                )
            )
        self.activedomain = self.padomains[0]
        self.prcnames = []

        for i in range(self.__rffile.read('i')[-1]):
            self.prcnames.append(self.__rffile.read('25s')[-1].strip())

        self.data_start_byte = self.__rffile.record_start
        self.record_fmt = self.id_fmt + str(len(self.prcnames)) + self.data_fmt
        self.record_size = self.__rffile.record_size
        self.SDATE, self.STIME, dummy, dummy, dummy, dummy, dummy, dummy = \
            self.__rffile.read(self.id_fmt)
        self.__rffile.previous()
        self.TSTEP = 100.
        self.padded_size = self.record_size + 8
        domain = self.padomains[0]
        domi = (domain['iend'] - domain['istart'] + 1)
        domj = (domain['jend'] - domain['jstart'] + 1)
        doml = (domain['tlay'] - domain['blay'] + 1)
        self.records_per_time = self.nspec * domi * domj * doml
        self.time_data_block = self.padded_size * self.records_per_time
        self.time_step = 100.

    def timerange(self):
        t1 = (self.start_date, self.start_time + self.time_step)
        t2 = timeadd((self.end_date, self.end_time), (0, self.time_step))
        return timerange(t1, t2, self.time_step)


class TestMemmap(unittest.TestCase):
    def runTest(self):
        pass

    def setUp(self):
        pass

    def testIPR(self):
        warn('Test not implemented; data too big for distribution')


if __name__ == '__main__':
    unittest.main()



__all__ = ['gcnc_base', 'gcnc']
from PseudoNetCDF.core._files import netcdf
from PseudoNetCDF import PseudoNetCDFFile
from PseudoNetCDF.pncwarn import warn
import numpy as np


class gcnc_base(PseudoNetCDFFile):
    def _newlike(self):
        if isinstance(self, PseudoNetCDFFile):
            outt = gcnc_base
            outf = outt.__new__(outt)
        else:
            outf = PseudoNetCDFFile()
        return outf

    def ll2xy(self, lon, lat, *args, **kwds):
        return lon, lat

    def ll2ij(self, lon, lat, bounds='error', clean='none'):
        """
        Converts lon/lat to 0-based indicies (0,M), (0,N)

        Parameters
        ----------
        lon : scalar or iterable of longitudes in decimal degrees
        lat : scalar or iterable of latitudes in decimal degrees
        bounds : ignore, error, warn if i,j are out of domain

        Returns
        -------
        i, j : indices (0-based) for variables
        """
        lon = np.asarray(lon)
        lat = np.asarray(lat)

        lonc = self.variables['lon']
        latc = self.variables['lat']

        dlon = np.median(np.diff(lonc)) / 2.
        dlat = np.median(np.diff(latc)) / 2.

        lonb = np.array([lonc - dlon, lonc + dlon]).T
        latb = np.array([latc - dlat, latc + dlat]).T

        spi = latb[:, 0] < -90
        npi = latb[:, 0] > 90

        latb[spi, 0] = latc[spi] - dlon / 2.
        latb[spi, 1] = latc[spi] + dlon / 2.
        latb[npi, 0] = latc[npi] - dlon / 2.
        latb[npi, 1] = latc[npi] + dlon / 2.

        easter = lon[:, None] >= lonb[:, 0]
        wester = lon[:, None] < lonb[:, 1]
        loni, gi = np.where(easter & wester)
        i = np.ma.masked_all(lon.shape, dtype='i')
        i[loni] = gi
        norther = lat[:, None] >= latb[:, 0]
        souther = lat[:, None] < latb[:, 1]
        latj, gj = np.where(norther & souther)
        j = np.ma.masked_all(lat.shape, dtype='i')
        j[latj] = gj

        nx = lonb.shape[0]
        ny = latb.shape[0]
        if bounds == 'ignore':
            pass
        else:
            missi, = np.where(i.mask)
            missj, = np.where(j.mask)
            lowi, = np.where(~easter[missi].any(1))
            highi, = np.where(~wester[missi].any(1))
            lowj, = np.where(~norther[missj].any(1))
            highj, = np.where(~souther[missj].any(1))
            outb = (lowi | lowj | highi | highj)
            nout = outb.sum()
            if nout > 0:
                message = '{} Points out of bounds; {}'.format(
                    nout, np.where(outb))
                if bounds == 'error':
                    raise ValueError(message)
                else:
                    warn(message)

        if clean == 'clip':
            i[lowi] = 0
            i[highi] = nx - 1
            j[lowj] = 0
            j[highj] = ny - 1
        else:
            # Mask or nothing should both create symmetric masks
            mask = (
                np.ma.getmaskarray(i) |
                np.ma.getmaskarray(j)
            )
            i = np.ma.masked_where(mask, i)
            j = np.ma.masked_where(mask, j)

        return i, j

    def interpSigma(self, vglvls, vgtop=None, interptype='linear',
                    extrapolate=False, fill_value='extrapolate',
                    layerdims=None,
                    approach='eta', verbose=0):
        """
        Parameters
        ----------
        self : the file to interpolate from must have VGLVLS
        vglvls : the new vglvls (edges)
        vgtop : Converting to new vgtop (Pascals)
        interptype : 'linear' or 'conserve'
             linear : uses a linear interpolation
             conserve : uses a mass conserving interpolation
        extrapolate : allow extrapolation beyond bounds with linear,
                      default False
        fill_value : set fill value (e.g, nan) to prevent extrapolation or edge
                     continuation
        layerdims : specify layer dimension, None will apply to all dimensions
                    named lev*
        approach :
             eta : use simple eta coordinates to calculate sigma and
                   interpolate
             pressure : requires surface pressure
        verbose : 0-inf show more

        Returns
        -------
        outf - ioapi_base PseudoNetCDFFile with al variables interpolated

        Notes
        -----
        When extrapolate is false, the edge values are used for points beyond
        the inputs.
        """
        etai = (
            self.variables['P0'][...] * self.variables['hybi'][...] +
            self.variables['hyai'][...]
        ) * 100.
        # grab input sigma coordinates
        myvglvls = (etai - vgtop) / (etai[0] - vgtop)

        # Use midpoint for sigmas in inputs and outputs
        zs = (myvglvls[:-1] + myvglvls[1:]) / 2.
        nzs = (vglvls[:-1] + vglvls[1:]) / 2.

        if interptype == 'linear':
            from ..coordutil import getinterpweights
            weights = getinterpweights(
                zs, nzs, kind=interptype, fill_value=fill_value,
                extrapolate=extrapolate)
            # Create a function for interpolation

            def interpsigma(data):
                if data.ndim == 1:
                    newdata = (weights[:data.shape[0]] * data[:, None]).sum(0)
                else:
                    newdata = (weights[None, :data.shape[0], :, None, None] *
                               data[:, :, None]).sum(1)
                return newdata
        else:
            raise ValueError(
                'interptype only implemented for "linear"; got ' + interptype
            )

        # Apply function on LAY
        if layerdims is None:
            layerdims = sorted([
                dk for dk in self.dimensions
                if (
                    dk.startswith('lev') and
                    not dk.endswith('_bounds') and
                    not dk == 'lev1')
            ])
            if verbose > 0:
                print(layerdims)
        dimfunc = dict([(layerkey, interpsigma) for layerkey in layerdims])
        outf = self.applyAlongDimensions(**dimfunc)

        return outf

    def stack(self, other, dimkey):
        from collections import Iterable
        outf = self._copywith(props=True, dimensions=False)
        if isinstance(other, Iterable):
            pfiles = other
        else:
            pfiles = [other]

        outf = PseudoNetCDFFile.stack(self, other, dimkey)
        tvar = outf.variables['time']
        tunits = tvar.units.strip()
        tres, rdatestr = tunits.split(' since ')

        if dimkey == 'time':
            times = self.getTimes()
            for pfile in pfiles:
                times = np.concatenate([times, pfile.getTimes()], axis=0)

        tformat = '%Y-%m-%d %H:%M:%S%z'
        rdate = times[0]
        rdatestr = rdate.strftime(tformat)
        tvar.units = 'hours since ' + rdatestr
        rvals = np.array([
            (t - rdate).total_seconds() / 3600. for t in times
        ])
        tvar[:] = rvals
        return outf


class gcnc(gcnc_base, netcdf):
    def __init__(self, *args, **kwds):
        netcdf.__init__(self, *args, **kwds)
        self.setCoords('hyam hyai hybm hybi time lat lon lev ilev P0'.split())



import PseudoNetCDF as pnc
import numpy as np

path='https://aura.gesdisc.eosdis.nasa.gov/opendap/hyrax/Aura_OMI_Level2/OMNO2.003/2020/001/OMI-Aura_L2-OMNO2_2020m0101t0117-o82246_v003-2020m0610t191058.he5'

key = 'CloudFraction'
satf = pnc.pncopen(path, format='netcdf')
#satf.set_auto_maskandscale(False)
subsetf = satf.subset([key])
copyf = pnc.PseudoNetCDFFile()
copyf.copyDimension(satf.dimensions['nTimes'], key='nTimes')
copyf.copyDimension(satf.dimensions['nXtrack'], key='nXtrack')
copyf.copyVariable(satf.variables[key], key=key)
print('Bad', np.unique(subsetf.variables[key][:]))
print('Good', np.unique(copyf.variables[key][:]))
print('Good', np.unique(satf.variables[key][:]))



__all__ = ['point_source']
__doc__ = """
.. _Read
:mod:`Read` -- point_source Read interface
============================================

.. module:: Read
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` random access read for CAMx
              point_source files.  See PseudoNetCDF.sci_var.PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""

# Distribution packages
import unittest
import struct

# Site-Packages
from numpy import zeros, array

# This Package modules
from PseudoNetCDF.camxfiles.timetuple import timediff, timerange
from PseudoNetCDF.camxfiles.FortranFileUtil import OpenRecordFile, read_into
from PseudoNetCDF.camxfiles.FortranFileUtil import Int2Asc
from PseudoNetCDF.sci_var import PseudoNetCDFFile, PseudoNetCDFVariables


# for use in identifying uncaught nan
listnan = struct.unpack('>f', b'\xff\xc0\x00\x00')[0]
checkarray = zeros((1,), 'f')
checkarray[0] = listnan
array_nan = checkarray[0]


class point_source(PseudoNetCDFFile):
    """
    point_source provides a PseudoNetCDF interface for CAMx
    point_source files.  Where possible, the inteface follows
    IOAPI conventions (see www.baronams.com).

    ex:
        >>> point_source_path = 'camx_point_source.bin'
        >>> rows,cols = 65,83
        >>> point_sourcefile = point_source(point_source_path,rows,cols)
        >>> point_sourcefile.variables.keys()
        ['TFLAG', 'ETFLAG', 'TFLAG', 'XSTK', 'YSTK', 'HSTK', 'DSTK', 'TSTK',
         'VSTK', 'KCELL', 'FLOW', 'PLMHT', 'NSTKS', 'NO', 'NO2', ...]
        >>> tflag = point_sourcefile.variables['TFLAG']
        >>> tflag.dimensions
        ('TSTEP', 'VAR', 'DATE-TIME')
        >>> tflag[0,0,:]
        array([2005185,       0])
        >>> tflag[-1,0,:]
        array([2005185,  240000])
        >>> v = point_sourcefile.variables['XSTK']
        >>> v.dimensions
        ('NSTK',)
        >>> v.shape
        (38452,)
        >>> v = point_sourcefile.variables['NO2']
        >>> v.dimensions
        ('TSTEP', 'NSTK')
        >>> v.shape
        (25, 38452)
        >>> point_sourcefile.dimensions
        {'TSTEP': 25, 'NSTK': 38452}
    """

    emiss_hdr_fmt = "10i60i3ifif"
    grid_hdr_fmt = "ffiffffiiiiifff"
    cell_hdr_fmt = "iiii"
    time_hdr_fmt = "ifif"
    spc_fmt = "10i"
    nstk_hdr_fmt = "ii"
    padded_nstk_hdr_size = struct.calcsize("ii" + nstk_hdr_fmt)
    padded_time_hdr_size = struct.calcsize("ii" + time_hdr_fmt)
    stk_hdr_fmt = "ffffff"
    id_fmt = "i" + spc_fmt
    id_size = struct.calcsize(id_fmt)
    data_fmt = "f"
    stkprops = ['XSTK', 'YSTK', 'HSTK', 'DSTK', 'TSTK', 'VSTK']
    stktimeprops = ['KCELL', 'FLOW', 'PLMHT']

    def __init__(self, rf):
        """
        Initialization included reading the header and learning
        about the format.

        see __readheader and __gettimestep() for more info
        """
        self.rffile = OpenRecordFile(rf)
        self.padded_time_hdr_size = struct.calcsize(self.time_hdr_fmt + "ii")
        self.__readheader()
        self.__gettimestep()
        self.__gettimeprops()
        self.createDimension('TSTEP', self.time_step_count)
        self.createDimension('STK', self.nstk)
        varkeys = (['XSTK', 'YSTK', 'HSTK', 'DSTK', 'TSTK', 'VSTK',
                    'KCELL', 'FLOW', 'PLMHT'] +
                   [i.strip() for i in self.spcnames])
        self.variables = PseudoNetCDFVariables(self.__var_get, varkeys)

    def __var_get(self, key):
        values = self.__variables(key)
        if key in self.stkprops:
            var = self.createVariable(key, 'f', ('STK',))
        else:
            var = self.createVariable(key, 'f', ('TSTEP', 'STK'))
        var[:] = values
        setattr(var, 'notread', 1)
        return var

    def __variables(self, k):
        if k in self.stkprops:
            return array(self.stk_props)[:, self.stkprops.index(k)]
        elif k in self.stktimeprops:
            stkps = array(self.stk_time_props)[:, :, 2:]
            return stkps[:, :, self.stktimeprops.index(k)]
        else:
            return self.getArray()[:, self.spcnames.index(k.ljust(10)), :]

    def header(self):
        rdum = 0.
        idum = 0
        ione = 1
        return [
            [self.name, self.note, ione, self.nspec, self.start_date,
                self.start_time, self.end_date, self.end_time],
            [rdum, rdum, self.iutm, self.xorg, self.yorg, self.delx, self.dely,
                self.nx, self.ny, self.nz, idum, idum, rdum, rdum, rdum],
            [ione, ione, self.nx, self.ny],
            self.spcnames,
            [ione, self.nstk],
            self.stk_props,
            self.stk_time_props
        ]

    def __readheader(self):
        """
        __readheader reads the header section of the ipr file
        it initializes each header field (see CAMx Users Manual for a list)
        as properties of the ipr class
        """
        vals = self.rffile.read(self.emiss_hdr_fmt)
        self.name = vals[0:10]
        self.note = vals[10:70]
        ione = vals[70]
        self.nspec = vals[71]
        self.start_date = vals[72]
        self.start_time = vals[73]
        self.end_date = vals[74]
        self.end_time = vals[75]

        vals = self.rffile.read(self.grid_hdr_fmt)
        rdum, rdum, self.iutm = vals[0:3]
        self.xorg, self.yorg, self.delx, self.dely = vals[3:7]
        self.nx, self.ny, self.nz = vals[7:10]
        idum, idum, rdum, rdum, rdum = vals[10:]
        if self.nz == 0:
            # Special case of gridded emissions
            # Seems to be same as avrg
            self.nlayers = 1
        else:
            self.nlayers = self.nz
        ione, ione, nx, ny = self.rffile.read(self.cell_hdr_fmt)
        if not (self.nx, self.ny) == (nx, ny):
            raise ValueError(("nx, ny defined first as %i, %i and then as " +
                              "%i, %i") % (self.nx, self.ny, nx, ny))
        species_temp = self.rffile.read(self.nspec * self.spc_fmt)
        self.spcnames = []
        for i in range(0, self.nspec * 10, 10):
            self.spcnames.append(Int2Asc(species_temp[i:i + 10]))

        ione, self.nstk = self.rffile.read(self.nstk_hdr_fmt)

        stkprms = zeros((self.nstk * len(self.stk_hdr_fmt),), 'f')
        read_into(self.rffile, stkprms, '')
        self.rffile.next()
        # self.rffile.previous()
        # self.tmplist=self.rffile.read('ffffff' * self.nstk)

        stkprms = stkprms.reshape((self.nstk, len(self.stk_hdr_fmt)))
        for i in range(stkprms.shape[0]):
            if stkprms[i, -1] == array_nan:
                stkprms[i, -1] = float('-nan')
        self.stk_props = stkprms.tolist()
        self.data_start_byte = self.rffile.record_start
        self.start_date, self.start_time, end_date, end_time = \
            self.rffile.read(self.time_hdr_fmt)

        self.time_step = timediff(
            (self.start_date, self.start_time), (end_date, end_time))
        # self.end_time += self.time_step
        mydayhrs = (2400, 24)[int(self.time_step % 2)]
        self.time_step_count = int(timediff((self.start_date, self.start_time),
                                            (self.end_date, self.end_time),
                                            mydayhrs) / self.time_step)

        self.stk_time_prop_fmt = "" + ("iiiff" * self.nstk)
        self.padded_stk_time_prop_size = struct.calcsize(
            "ii" + self.stk_time_prop_fmt)

        self.record_fmt = ("i10i") + self.data_fmt * (self.nstk)
        self.record_size = struct.calcsize(self.record_fmt)
        self.padded_size = self.record_size + 8

    def __gettimestep(self):
        """
        this is taken care of in the readheader routine
        record format provides start and end for each hour,
        which translates to t1 and t2
        """
        pass

    def __gettimeprops(self):
        self.stk_time_props = []
        dates = timerange((self.start_date, self.start_time),
                          (self.end_date, self.end_time),
                          self.time_step,
                          (2400, 24)[int(self.time_step % 2)])
        for ti, (d, t) in enumerate(dates):
            tmpprop = zeros((len(self.stk_time_prop_fmt)), 'f')
            tmpprop[...] = self.seekandread(
                d, t, 1, True, self.stk_time_prop_fmt)
            tmpprop = tmpprop.reshape(self.nstk, 5)
            for i in range(tmpprop.shape[0]):
                if tmpprop[i, -2] == array_nan:
                    tmpprop[i, -2] = float('-nan')

            self.stk_time_props.append(tmpprop.tolist())

    def __timerecords(self, dt):
        """
        Calculate the number of records to increment to reach time (d,t)
        """
        d, t = dt
        nsteps = int(timediff((self.start_date, self.start_time),
                              (d, t), (2400, 24)[int(self.time_step % 2)]))
        nspec = self.__spcrecords(self.nspec + 1)
        return nsteps * (nspec)

    def __spcrecords(self, spc):
        """
        Calculated number of records before spc
        """

        return spc - 1

    def __recordposition(self, date, time, spc, offset=False):
        """
        Use time (d,t), spc, and k to calculate number of records before
        desired record

        date - integer julian
        time - float
        spc - integer
        """
        ntime = self.__timerecords((date, time))
        nhdr = ((ntime // self.__spcrecords(self.nspec + 1)) + 1)
        nspc = self.__spcrecords(spc)
        noffset = -abs(int(offset))
        byte = self.data_start_byte
        byte += nhdr * (self.padded_time_hdr_size +
                        self.padded_nstk_hdr_size +
                        self.padded_stk_time_prop_size)
        byte += (ntime + nspc) * self.padded_size
        byte += noffset * self.padded_stk_time_prop_size
        return byte

    def seek(self, date=None, time=None, spc=-1, offset=False):
        """
        Move file cursor to the beginning of the specified record
        see __recordposition for parameter definitions
        """
        seekto = self.__recordposition(date, time, spc, offset)
        self.rffile._newrecord(seekto)

    def read(self, fmt=None):
        """
        Provide direct access to record file read
        """
        if fmt is None:
            fmt = self.record_fmt
        return self.rffile.read(fmt)

    def read_into(self, dest):
        """
        Transfer values from current record to dest
        dest - numeric or numpy array
        """

        return read_into(self.rffile, dest, self.id_fmt, self.data_fmt)

    def seekandreadinto(self, dest, date=None, time=None, spc=1):
        """
        see seek and read_into
        """

        self.seek(date, time, spc)
        self.read_into(dest)

    def seekandread(self, date=None, time=None, spc=1, offset=False, fmt=None):
        """
        see seek and read
        """
        self.seek(date, time, spc, offset)
        return self.read(fmt)

    def values(self):
        for d, t, spc in self.__iter__():
            yield self.seekandread(d, t, spc)

    def items(self):
        for d, t, spc in self.__iter__():
            yield d, t, spc, self.seekandread(d, t, spc)

    def keys(self):
        for ti, (d, t) in enumerate(self.timerange()):
            for spc in range(1, len(self.spcnames) + 1):
                yield d, t, spc

    __iter__ = keys

    def getArray(self):
        a = zeros((self.time_step_count, self.nspec, self.nstk), 'f')
        for ti, (d, t) in enumerate(self.timerange()):
            for spc in range(1, len(self.spcnames) + 1):
                self.seekandreadinto(a[ti, spc - 1, ...], d, t, spc)
        return a.copy()

    def timerange(self):
        return timerange((self.start_date, self.start_time),
                         (self.end_date, self.end_time), self.time_step,
                         eod=24)


class TestRead(unittest.TestCase):
    def runTest(self):
        pass

    def setUp(self):
        pass

    def testPT(self):
        import PseudoNetCDF.testcase
        emissfile = point_source(
            PseudoNetCDF.testcase.camxfiles_paths['point_source'])
        v = emissfile.variables['NO2']
        self.assertTrue((v[:] == array(
            [0.00000000e+00, 3.12931000e+02, 1.23599997e+01, 0.00000000e+00,
             5.27999992e+01, 0.00000000e+00, 3.12931000e+02, 1.23599997e+01,
             0.00000000e+00, 5.27999992e+01], dtype='f').reshape(2, 5)).all())


if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np


np_all_close = np.testing.assert_allclose


class IOAPITest(unittest.TestCase):
    def setUp(self):
        from datetime import datetime, timedelta
        self.testncf = self._makencf()
        self.mymeta = set([
            'time', 'time_bounds', 'latitude', 'longitude', 'latitude_bounds',
            'longitude_bounds', 'lambert_conformal_conic'
        ])
        self.myvars = self.mymeta.union(['O3'])
        self.mydims = ['TIME', 'LAY', 'ROW', 'COL', 'nv', 'tnv']
        rtime = datetime.strptime(
            '1970-01-01 00:00:00+0000', '%Y-%m-%d %H:%M:%S%z')
        self.mytimes = np.array([rtime + timedelta(hours=i)
                                 for i in range(24)])
        self.testf = self._makencf()

    def _makencf(self):
        from PseudoNetCDF.cmaqfiles import ioapi_base
        minef = ioapi_base()

        minef.createDimension('TSTEP', 24)
        minef.createDimension('DATE-TIME', 2)
        minef.createDimension('LAY', 3)
        minef.createDimension('VAR', 2)
        minef.createDimension('ROW', 5)
        minef.createDimension('COL', 6)

        minef.EXEC_ID = "mcip ".ljust(80)
        minef.FTYPE = 1
        minef.CDATE = 2017069
        minef.CTIME = 144432
        minef.WDATE = 2017069
        minef.WTIME = 144432
        minef.SDATE = 2011001
        minef.STIME = 0
        minef.TSTEP = 10000
        minef.NTHIK = 1
        minef.NCOLS = 6
        minef.NROWS = 5
        minef.NLAYS = 3
        minef.NVARS = 2
        minef.GDTYP = 2
        minef.P_ALP = 33.
        minef.P_BET = 45.
        minef.P_GAM = -97.
        minef.XCENT = -97.
        minef.YCENT = 40.
        minef.XORIG = -2736000.
        minef.YORIG = -2088000.
        minef.XCELL = 36000.
        minef.YCELL = 36000.
        minef.VGTYP = 7
        minef.VGTOP = 5000.
        minef.VGLVLS = np.array([1., 0.9975, 0.995, 0.9925], dtype='f')
        minef.GDNAM = "METCRO_36US1_CRO"
        minef.UPNAM = "METCRO          "

        o3 = minef.createVariable(
            'O3', 'f', ('TSTEP', 'LAY', 'ROW', 'COL'), units='ppbV'
        )
        no2 = minef.createVariable(
            'NO2', 'f', ('TSTEP', 'LAY', 'ROW', 'COL'), units='ppbV'
        )

        o3[:] = np.arange(24 * 3 * 5 * 6).reshape(24, 3, 5, 6)
        no2[:] = np.arange(24 * 3 * 5 * 6).reshape(24, 3, 5, 6) * .05
        return minef

    def testTFLAG(self):
        testf = self.testf.copy()
        testf.updatetflag()
        test = self.assertEqual
        test(2011001, testf.variables['TFLAG'][0, 0, 0])
        test(0, testf.variables['TFLAG'][0, 0, 1])
        test(testf.SDATE, testf.variables['TFLAG'][0, 0, 0])
        test(testf.STIME, testf.variables['TFLAG'][0, 0, 1])
        test(testf.NVARS, testf.variables['TFLAG'].shape[1])

    def testSubsetVariables(self):
        testf = self.testf.copy()
        testf.updatetflag()
        test = self.assertEqual
        outf = testf.subsetVariables(['O3'])
        test(1, outf.variables['TFLAG'].shape[1])
        test(outf.NVARS, outf.variables['TFLAG'].shape[1])
        test('O3'.ljust(16), getattr(outf, 'VAR-LIST'))

    def _trimdim(self, dim):
        reff = self.testf.copy()
        reff.updatetflag()
        newf = reff.sliceDimensions(**{dim: slice(1, -1)})
        test = self.assertEqual
        attrkey = 'N{}S'.format(dim)
        test(getattr(reff, attrkey) - 2, getattr(newf, attrkey))
        test(len(newf.dimensions[dim]), getattr(newf, attrkey))
        return reff, newf

    def testSliceRow(self):
        reff, newf = self._trimdim('ROW')
        self.assertEqual(newf.YORIG, reff.YORIG + reff.YCELL)

    def testSliceCol(self):
        reff, newf = self._trimdim('COL')
        self.assertEqual(newf.XORIG, reff.XORIG + reff.XCELL)

    def testSliceLay(self):
        reff, newf = self._trimdim('LAY')
        self.assertEqual(True, np.allclose(newf.VGLVLS, reff.VGLVLS[1:-1]))

    def _meandim(self, dim):
        reff = self.testf.copy()
        reff.updatetflag()
        newf = reff.applyAlongDimensions(**{dim: 'mean'})
        test = self.assertEqual
        attrkey = 'N{}S'.format(dim)
        test(1, getattr(newf, attrkey))
        test(len(newf.dimensions[dim]), getattr(newf, attrkey))
        return reff, newf

    def testMeanRow(self):
        reff, newf = self._meandim('ROW')

    def testMeanCol(self):
        reff, newf = self._meandim('COL')

    def testMeanLay(self):
        reff, newf = self._meandim('LAY')
        test = self.assertEqual
        newl = reff.VGLVLS[:-1].mean()
        newu = reff.VGLVLS[1:].mean()
        test(True, np.allclose([newl, newu], newf.VGLVLS))

    def testInterpSigma(self):
        reff = self.testf.copy()
        reff.updatetflag()
        # Simple resample
        newf = reff.interpSigma(
            reff.VGLVLS[1:], vgtop=reff.VGTOP, interptype='conserve'
        )
        test = np.allclose(
            reff.variables['O3'][:, 1:], newf.variables['O3'][:]
        )
        self.assertEqual(True, test)
        newf = reff.interpSigma(
            np.convolve([.5, .5], reff.VGLVLS, mode='valid'),
            vgtop=reff.VGTOP, interptype='conserve'
        )
        test = np.allclose(
            0.5 * reff.variables['O3'][:, 1:] +
            0.5 * reff.variables['O3'][:, :-1],
            newf.variables['O3'][:]
        )


class ProfileTest(unittest.TestCase):
    def setUp(self):
        pass

    def testIcon(self):
        from ..cmaqfiles import icon_profile
        from ..testcase import cmaqfiles_paths
        icon_file = icon_profile(cmaqfiles_paths['icon_profile'])
        itestvals = {
            "O3": [
                3.5E-02, 3.5E-02, 4.E-02, 5.E-02, 6.E-02, 7.E-02],
            "ASO4I": [
                4.810E-03, 4.810E-03, 3.207E-03, 3.207E-03, 6.413E-04,
                3.207E-04
            ],
            "NUMATKN": [
                1.478E+09, 1.470E+09, 9.774E+08, 9.655E+08, 1.920E+08,
                9.584E+07
            ],
            "SRFACC": [
                1.491E-05, 1.373E-05, 8.694E-06, 6.815E-06, 1.190E-06,
                5.697E-07
            ]
        }
        test = self.assertEqual
        for testk, testvals in itestvals.items():
            vals = icon_file.variables[testk]
            test(True, np.allclose(vals, testvals))

    def testBcon(self):
        from ..cmaqfiles import bcon_profile
        from ..testcase import cmaqfiles_paths
        btestvals = {
            "O3": [
                3.000E-02, 3.500E-02, 4.000E-02, 5.000E-02, 6.000E-02,
                7.000E-02, 3.000E-02, 3.500E-02, 4.000E-02, 5.000E-02,
                6.000E-02, 7.000E-02, 3.500E-02, 3.500E-02, 4.000E-02,
                5.000E-02, 6.000E-02, 7.000E-02, 3.500E-02, 4.000E-02,
                4.500E-02, 5.000E-02, 6.000E-02, 7.000E-02
            ],
            "ASO4I": [
                6.413E-03, 6.413E-03, 6.413E-03, 3.207E-03, 6.413E-04,
                3.207E-04, 6.413E-03, 6.413E-03, 6.413E-03, 6.413E-03,
                6.413E-04, 3.207E-04, 4.810E-03, 4.810E-03, 3.207E-03,
                3.207E-03, 6.413E-04, 3.207E-04, 9.620E-03, 6.413E-03,
                6.413E-03, 3.207E-03, 6.413E-04, 3.207E-04
            ],
            "NUMATKN": [
                1.957E+09, 1.949E+09, 1.936E+09, 9.655E+08, 1.920E+08,
                9.584E+07, 1.957E+09, 1.949E+09, 1.936E+09, 1.924E+09,
                1.920E+08, 9.584E+07, 1.478E+09, 1.470E+09, 9.774E+08,
                9.655E+08, 1.920E+08, 9.584E+07, 2.915E+09, 1.949E+09,
                1.936E+09, 9.655E+08, 1.920E+08, 9.584E+07
            ],
            "SRFACC": [
                1.776E-05, 1.658E-05, 1.439E-05, 6.815E-06, 1.190E-06,
                5.697E-07, 1.776E-05, 1.658E-05, 1.439E-05, 1.251E-05,
                1.190E-06, 5.697E-07, 1.491E-05, 1.373E-05, 8.694E-06,
                6.815E-06, 1.190E-06, 5.697E-07, 2.346E-05, 1.658E-05,
                1.439E-05, 6.815E-06, 1.190E-06, 5.697E-07
            ],
        }
        bcon_file = bcon_profile(cmaqfiles_paths['bcon_profile'])
        test = self.assertEqual
        for testk, testvals in btestvals.items():
            vals = bcon_file.variables[testk]
            test(True, np.allclose(vals[:].T.ravel(), testvals))



__all__ = ['temperature']
__doc__ = """
.. _Memmap
:mod:`Memmap` -- temperature Memmap interface
=============================================

.. module:: Memmap
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` memory map for CAMx
              temperature files.  See PseudoNetCDF.sci_var.PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""

# Distribution packages
import unittest
import struct

# Site-Packages
from numpy import zeros, array, memmap

# This Package modules
from PseudoNetCDF.sci_var import PseudoNetCDFFile, PseudoNetCDFVariable
from PseudoNetCDF.sci_var import PseudoNetCDFVariables
from PseudoNetCDF.ArrayTransforms import ConvertCAMxTime

# for use in identifying uncaught nan
listnan = struct.unpack('>f', b'\xff\xc0\x00\x00')[0]
checkarray = zeros((1,), 'f')
checkarray[0] = listnan
array_nan = checkarray[0]


class temperature(PseudoNetCDFFile):
    """
    temperature provides a PseudoNetCDF interface for CAMx
    temperature files.  Where possible, the inteface follows
    IOAPI conventions (see www.baronams.com).

    ex:
        >>> temperature_path = 'camx_temperature.bin'
        >>> rows,cols = 65,83
        >>> temperaturefile = temperature(temperature_path,rows,cols)
        >>> temperaturefile.variables.keys()
        ['TFLAG', 'AIRTEMP', 'SURFTEMP']
        >>> tflag = temperaturefile.variables['TFLAG']
        >>> tflag.dimensions
        ('TSTEP', 'VAR', 'DATE-TIME')
        >>> tflag[0,0,:]
        array([2005185,       0])
        >>> tflag[-1,0,:]
        array([2005185,  240000])
        >>> v = temperaturefile.variables['SURFTEMP']
        >>> v.dimensions
        ('TSTEP', 'ROW', 'COL')
        >>> v.shape
        (25, 65, 83)
        >>> v = temperaturefile.variables['AIRTEMP']
        >>> v.dimensions
        ('TSTEP', 'LAY', 'ROW', 'COL')
        >>> v.shape
        (25, 28, 65, 83)
        >>> temperaturefile.dimensions
        {'TSTEP': 25, 'LAY': 28, 'ROW': 65, 'COL': 83}
    """

    id_fmt = 'fi'
    data_fmt = 'f'

    def __init__(self, rf, rows=None, cols=None):
        self.__memmap = memmap(rf, '>f', 'r', offset=0)

        rowsXcols = self.__memmap[0].view('i') // 4 - 2
        record_length = rowsXcols + 4
        records = self.__memmap.size // record_length

        times = self.__memmap.reshape(records, record_length)[:, 1:3]
        self.STIME, self.SDATE = times[0]
        for i, (t, d) in enumerate(times):
            if (t, d) != (self.STIME, self.SDATE):
                break
        self.SDATE = self.SDATE.view('i')
        self.createDimension('LAY', i - 1)
        self.createDimension('TSTEP', times.shape[0] / i)

        if rows is None and cols is None:
            rows = rowsXcols
            cols = 1
        elif rows is None:
            rows = rowsXcols / cols
        elif cols is None:
            cols = rowsXcols / rows
        else:
            if cols * rows != rowsXcols:
                raise ValueError(("The product of cols (%d) and rows (%d) " +
                                  "must equal cells (%d)") % (
                                 cols, rows, rowsXcols))

        self.createDimension('ROW', rows)
        self.createDimension('COL', cols)
        self.createDimension('DATE-TIME', 2)
        self.createDimension('VAR', 2)

        self.NVARS = len(self.dimensions['VAR'])
        self.NLAYS = len(self.dimensions['LAY'])
        self.NROWS = len(self.dimensions['ROW'])
        self.NCOLS = len(self.dimensions['COL'])
        self.FTYPE = 1

        self.variables = PseudoNetCDFVariables(
            self.__var_get, ['AIRTEMP', 'SURFTEMP', 'TFLAG'])

    def __var_get(self, key):
        lays = len(self.dimensions['LAY'])
        times = len(self.dimensions['TSTEP'])
        rows = len(self.dimensions['ROW'])
        cols = len(self.dimensions['COL'])
        surf = 1
        air = 2
        time = 3
        date = 4
        out_idx = zeros(self.__memmap.shape, dtype='b').reshape(
            times, lays + 1, rows * cols + 4)
        out_idx[:, 0, 3:-1] = surf
        out_idx[:, 1:, 3:-1] = air
        out_idx[:, :, 1] = time
        out_idx[:, :, 2] = date
        out_idx = out_idx.ravel()
        buf = self.__memmap[out_idx == 0].reshape((lays + 1) * times, 2)
        if not (buf[:, 0] == buf[:, 1]).all():
            raise ValueError("Buffer")
        tmpvals = self.__memmap[out_idx == 1].reshape(times, rows, cols)
        v = PseudoNetCDFVariable(self, 'SURFTEMP', 'f',
                                 ('TSTEP', 'ROW', 'COL'),
                                 values=tmpvals)
        self.variables['SURFTEMP'] = v
        v.units = 'K'
        v.long_name = 'SURFTEMP'
        v.var_desc = 'SURFTEMP'
        tmpvals = self.__memmap[out_idx == 2].reshape(times, lays, rows, cols)
        v = PseudoNetCDFVariable(self, 'AIRTEMP', 'f',
                                 ('TSTEP', 'LAY', 'ROW', 'COL'),
                                 values=tmpvals)
        self.variables['AIRTEMP'] = v
        v.units = 'K'
        v.long_name = 'AIRTEMP'
        v.var_desc = 'AIRTEMP'

        date = self.__memmap[out_idx == date].view('>i')[0:None:lays + 1]
        time = self.__memmap[out_idx == time].view('>f')[0:None:lays + 1]
        tmpvals = ConvertCAMxTime(date, time, 2)
        v = PseudoNetCDFVariable(self, 'TFLAG', 'f',
                                 ('TSTEP', 'VAR', 'DATE-TIME'),
                                 values=tmpvals)
        self.variables['TFLAG'] = v

        return self.variables[key]


class TestMemmap(unittest.TestCase):
    def runTest(self):
        pass

    def setUp(self):
        pass

    def testTEMP(self):
        import PseudoNetCDF.testcase
        tempfile = temperature(
            PseudoNetCDF.testcase.camxfiles_paths['temperature'], 4, 5)
        tempfile.variables['TFLAG']
        checkat = array([2.97762360e+02, 2.97261993e+02, 3.00761200e+02,
                         3.03811005e+02, 3.04561218e+02, 2.96350311e+02,
                         2.96676544e+02, 3.00992096e+02, 3.05474762e+02,
                         3.07840637e+02, 2.99522430e+02, 3.00271698e+02,
                         3.03738403e+02, 3.07201843e+02, 3.08288422e+02,
                         3.02957214e+02, 3.04927643e+02, 3.06630157e+02,
                         3.07726074e+02, 3.07380707e+02, 2.97516449e+02,
                         2.96920105e+02, 3.00340576e+02, 3.03413177e+02,
                         3.04202728e+02, 2.96074036e+02, 2.96250641e+02,
                         3.00632294e+02, 3.05113647e+02, 3.07390533e+02,
                         2.99310059e+02, 2.99901031e+02, 3.03344666e+02,
                         3.06782135e+02, 3.07819946e+02, 3.02657013e+02,
                         3.04522675e+02, 3.06167206e+02, 3.07235107e+02,
                         3.06883484e+02, 2.97677338e+02, 2.96919098e+02,
                         3.00031250e+02, 3.03082672e+02, 3.03850861e+02,
                         2.96460999e+02, 2.95947815e+02, 3.00303680e+02,
                         3.04781982e+02, 3.07048492e+02, 2.99246979e+02,
                         2.99508667e+02, 3.02997650e+02, 3.06450500e+02,
                         3.07478485e+02, 3.02246765e+02, 3.04192139e+02,
                         3.05832489e+02, 3.06897644e+02, 3.06546173e+02,
                         2.97428253e+02, 2.97174896e+02, 3.00208191e+02,
                         3.03096893e+02, 3.04174133e+02, 2.96558685e+02,
                         2.96706177e+02, 3.00862610e+02, 3.04807037e+02,
                         3.06937347e+02, 2.98850220e+02, 2.99482727e+02,
                         3.03085022e+02, 3.06456787e+02, 3.07406586e+02,
                         3.01888580e+02, 3.03996735e+02, 3.05916962e+02,
                         3.07113647e+02, 3.06539337e+02, 2.97645966e+02,
                         2.97326630e+02, 3.00117950e+02, 3.02804077e+02,
                         3.03801544e+02, 2.96783661e+02, 2.96694946e+02,
                         3.00722931e+02, 3.04501587e+02, 3.06560150e+02,
                         2.98854828e+02, 2.99314972e+02, 3.02861023e+02,
                         3.06150177e+02, 3.07073944e+02, 3.01700745e+02,
                         3.03746124e+02, 3.05626617e+02, 3.06770447e+02,
                         3.06172394e+02, 2.97927094e+02, 2.97691681e+02,
                         3.00104675e+02, 3.02464874e+02, 3.03398926e+02,
                         2.97336578e+02, 2.97074127e+02, 3.00716736e+02,
                         3.04132446e+02, 3.06129700e+02, 2.98817017e+02,
                         2.99221039e+02, 3.02649536e+02, 3.05787415e+02,
                         3.06698334e+02, 3.01333618e+02, 3.03411346e+02,
                         3.05317505e+02, 3.06446869e+02,
                         3.05815948e+02], dtype='f').reshape(2, 3, 4, 5)
        self.assertTrue((tempfile.variables['AIRTEMP'] == checkat).all())

    def testNCF2TEMP(self):
        import PseudoNetCDF.testcase
        from PseudoNetCDF.pncgen import pncgen
        import os
        inpath = PseudoNetCDF.testcase.camxfiles_paths['temperature']
        outpath = inpath + '.check'
        infile = temperature(inpath, 4, 5)
        pncgen(infile, outpath, format='camxfiles.temperature')
        orig = open(inpath, 'rb').read()
        new = open(outpath, 'rb').read()
        assert(orig == new)
        os.remove(outpath)


if __name__ == '__main__':
    unittest.main()



__all__ = ['one3d']
__doc__ = """
.. _Read
:mod:`Read` -- one3d Read interface
============================================

.. module:: Read
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` random access read for CAMx
              generic 1 3D variable wind files.  See
              PseudoNetCDF.sci_var.PseudoNetCDFFile for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""
# Distribution packages
import unittest
import struct

# Site-Packages
from numpy import zeros, array

# This Package modules
from PseudoNetCDF.camxfiles.timetuple import timediff, timeadd, timerange
from PseudoNetCDF.camxfiles.FortranFileUtil import OpenRecordFile, read_into
from PseudoNetCDF.sci_var import PseudoNetCDFFile, PseudoNetCDFVariables

# for use in identifying uncaught nan
listnan = struct.unpack('>f', b'\xff\xc0\x00\x00')[0]
checkarray = zeros((1,), 'f')
checkarray[0] = listnan
array_nan = checkarray[0]


class one3d(PseudoNetCDFFile):
    """
    one3d provides a PseudoNetCDF interface for CAMx
    one3d files.  Where possible, the inteface follows
    IOAPI conventions (see www.baronams.com).

    ex:
        >>> one3d_path = 'camx_one3d.bin'
        >>> rows,cols = 65,83
        >>> one3dfile = one3d(one3d_path,rows,cols)
        >>> one3dfile.variables.keys()
        ['TFLAG', 'UNKNOWN']
        >>> tflag = one3dfile.variables['TFLAG']
        >>> tflag.dimensions
        ('TSTEP', 'VAR', 'DATE-TIME')
        >>> tflag[0,0,:]
        array([2005185,       0])
        >>> tflag[-1,0,:]
        array([2005185,  240000])
        >>> v = one3dfile.variables['UNKNOWN']
        >>> v.dimensions
        ('TSTEP', 'LAY', 'ROW', 'COL')
        >>> v.shape
        (25, 28, 65, 83)
        >>> one3dfile.dimensions
        {'TSTEP': 25, 'LAY': 28, 'ROW': 65, 'COL': 83}
    """

    id_fmt = "fi"
    data_fmt = "f"
    var_name = "UNKNOWN"
    units = "UNKNOWN"

    def __init__(self, rf, rows=None, cols=None):
        """
        Initialization included reading the header and learning
        about the format.

        see __readheader and __gettimestep() for more info
        """

        self.rffile = OpenRecordFile(rf)

        self.id_size = struct.calcsize(self.id_fmt)
        self.__readheader()
        self.__gettimestep()
        if rows is None and cols is None:
            rows = self.cell_count
            cols = 1
        elif rows is None:
            rows = self.cell_count / cols
        elif cols is None:
            cols = self.cell_count / rows
        else:
            if cols * rows != self.cell_count:
                raise ValueError(("The product of cols (%d) and rows (%d) " +
                                  "must equal cells (%d)") % (cols, rows,
                                                              self.cell_count))

        self.createDimension('TSTEP', self.time_step_count)
        self.createDimension('COL', cols)
        self.createDimension('ROW', rows)
        self.createDimension('LAY', self.nlayers)

        self.variables = PseudoNetCDFVariables(self.__var_get, [self.var_name])

    def __var_get(self, key):
        props = dict(units=self.units, var_desc=self.var_name.ljust(16),
                     long_name=self.var_name.ljust(16))
        values = self.getArray()

        var = self.createVariable(key, 'f', ('TSTEP', 'LAY', 'ROW', 'COL'))
        var[:] = values
        for k, v in props.items():
            setattr(var, k, v)
        return var

    def __readheader(self):
        """
        __readheader reads the header section of the vertical diffusivity file
        it initializes each header field (see CAMx Users Manual for a list)
        as properties of the ipr class
        """
        self.data_start_byte = 0
        self.start_time, self.start_date = self.rffile.read(self.id_fmt)
        self.record_size = self.rffile.record_size
        self.padded_size = self.record_size + 8
        self.cell_count = (
            self.record_size - self.id_size) // struct.calcsize(self.data_fmt)
        self.record_fmt = self.id_fmt + self.data_fmt * (self.cell_count)

    def __gettimestep(self):
        """
        Header information provides start and end date, but does not
        indicate the increment between.  This routine reads the first
        and second date/time and initializes variables indicating the
        timestep length and the anticipated number.
        """
        self.rffile._newrecord(
            self.padded_size
        )
        d, t = self.start_date, self.start_time
        self.nlayers = 0
        while timediff((self.start_date, self.start_time), (d, t)) == 0:
            t, d = self.rffile.read(self.id_fmt)
            self.nlayers += 1
        self.time_step = timediff((self.start_date, self.start_time), (d, t))

        while True:
            try:
                self.seek(d, t, 1, False)
                d, t = timeadd((d, t), (0, self.time_step))
            except Exception:
                break
        self.end_date, self.end_time = timeadd((d, t), (0, -self.time_step))
        self.time_step_count = int(timediff(
            (self.start_date, self.start_time),
            (self.end_date, self.end_time)) / self.time_step) + 1

    def __timerecords(self, dt):
        """
        routine returns the number of records to increment from the
        data start byte to find the first time
        """
        d, t = dt
        nsteps = int(
            timediff((self.start_date, self.start_time),
                     (d, t)) / self.time_step)
        nk = self.__layerrecords(self.nlayers + 1)
        return nsteps * nk

    def __layerrecords(self, k):
        """
        routine returns the number of records to increment from the
        data start byte to find the first klayer
        """
        return k - 1

    def __recordposition(self, date, time, k):
        """
        routine uses timerecords and layerrecords multiplied
        by the fortran padded size to return the byte position
        of the specified record

        date - integer
        time - float
        k - integer
        """
        ntime = self.__timerecords((date, time))
        nk = self.__layerrecords(k)
        return (nk + ntime) * self.padded_size + self.data_start_byte

    def seek(self, date=None, time=None, k=1, chkvar=True):
        """
        Move file cursor to beginning of specified record
        see __recordposition for a definition of variables
        """
        if date is None:
            date = self.start_date
        if time is None:
            time = self.start_time

        if chkvar:
            chkt1 = timediff((self.end_date, self.end_time), (date, time)) > 0
            chkt2 = timediff((self.start_date, self.start_time),
                             (date, time)) < 0
            if chkt1 or chkt2:
                raise KeyError(("Vertical Diffusivity file includes " +
                                "(%i,%6.1f) thru (%i,%6.1f); you requested " +
                                "(%i,%6.1f)") % (self.start_date,
                                                 self.start_time,
                                                 self.end_date,
                                                 self.end_time, date, time))
            if k < 1 or k > self.nlayers:
                raise KeyError(("Vertical Diffusivity file include layers 1" +
                                "thru %i; you requested %i") % (self.nlayers,
                                                                k))
        self.rffile._newrecord(self.__recordposition(date, time, k))

    def read(self):
        """
        provide direct access to the underlying RecordFile read
        method
        """
        return self.rffile.read(self.record_fmt)

    def read_into(self, dest):
        """
        put values from rffile read into dest
        dest - numpy or numeric array
        """
        return read_into(self.rffile, dest, self.id_fmt, self.data_fmt)

    def seekandreadinto(self, dest, date=None, time=None, k=1):
        """
        see seek and read_into
        """
        self.seek(date, time, k)
        return self.read_into(dest)

    def seekandread(self, date=None, time=None, k=1):
        """
        see seek and read
        """
        self.seek(date, time, k)
        return self.read()

    def values(self):
        for d, t, k in self.__iter__():
            yield self.seekandread(d, t, k)

    def items(self):
        for d, t, k in self.__iter__():
            yield d, t, k, self.seekandread(d, t, k)

    def keys(self):
        for d, t in self.timerange():
            for k in range(1, self.nlayers + 1):
                yield d, t, k
    __iter__ = keys

    def getArray(self):
        a = zeros((self.time_step_count, self.nlayers, len(
            self.dimensions['ROW']), len(self.dimensions['COL'])), 'f')
        for ti, (d, t) in enumerate(self.timerange()):
            for ki, k in enumerate(range(1, self.nlayers + 1)):
                self.seekandreadinto(a[ti, ki, ...], d, t, k)
        return a

    def timerange(self):
        return timerange((self.start_date, self.start_time),
                         (self.end_date, self.end_time + self.time_step),
                         self.time_step)


class TestRead(unittest.TestCase):
    def runTest(self):
        pass

    def setUp(self):
        pass

    def testKV(self):
        import PseudoNetCDF.testcase
        inpath = PseudoNetCDF.testcase.camxfiles_paths['vertical_diffusivity']
        vdfile = one3d(inpath, 4, 5)
        checkv = array([1.00000000e+00, 4.76359320e+00, 1.92715893e+01,
                        1.52158489e+01, 7.20601225e+00, 1.84097159e+00,
                        2.63084507e+01, 1.27621298e+01, 1.06348248e+01,
                        2.22587357e+01, 1.00000000e+00, 1.69009724e+01,
                        1.00000000e+00, 2.23075104e+01, 1.27485418e+01,
                        1.45508013e+01, 1.45637455e+01, 2.95294094e+01,
                        3.02676849e+01, 2.84974957e+01, 1.00000000e+00,
                        6.93706131e+00, 3.07418957e+01, 3.41300621e+01,
                        1.66266994e+01, 1.00000000e+00, 2.53920174e+01,
                        1.92539787e+01, 2.32906532e+01, 5.96702042e+01,
                        1.00000000e+00, 2.24458847e+01, 1.00000000e+00,
                        5.45038452e+01, 3.45825729e+01, 3.43578224e+01,
                        3.76071548e+01, 6.76799850e+01, 7.33648529e+01,
                        7.35801239e+01, 1.00000000e+00, 6.38178444e+00,
                        4.39327278e+01, 5.00754166e+01, 2.44474106e+01,
                        1.00000000e+00, 3.51700935e+01, 2.71137428e+01,
                        3.40312347e+01, 8.85775909e+01, 1.00000000e+00,
                        3.13994522e+01, 1.00000000e+00, 8.07169266e+01,
                        5.12892876e+01, 5.05734329e+01, 5.56603966e+01,
                        1.00394188e+02, 1.08980370e+02, 1.09251083e+02,
                        1.00000000e+00, 1.00000000e+00, 1.00000000e+00,
                        1.00000000e+00, 1.00000000e+00, 1.00000000e+00,
                        1.00000000e+00, 1.00000000e+00, 1.00000000e+00,
                        1.64916098e+00, 1.00000000e+00, 1.00000000e+00,
                        1.00000000e+00, 1.22205174e+00, 1.00000000e+00,
                        1.00000000e+00, 1.00000000e+00, 1.05408611e+01,
                        1.26687222e+01, 1.21386652e+01, 1.00000000e+00,
                        1.00000000e+00, 1.00000000e+00, 1.00000000e+00,
                        1.00000000e+00, 1.00000000e+00, 1.00000000e+00,
                        1.00000000e+00, 1.00000000e+00, 1.93040049e+00,
                        1.00000000e+00, 1.00000000e+00, 1.00000000e+00,
                        1.39350021e+00, 1.02697349e+00, 1.00000000e+00,
                        1.00000000e+00, 1.82250175e+01, 2.90407104e+01,
                        2.83827496e+01, 1.00000000e+00, 1.00000000e+00,
                        1.00000000e+00, 1.00000000e+00, 1.00000000e+00,
                        1.00000000e+00, 1.00000000e+00, 1.00000000e+00,
                        1.00000000e+00, 2.02609706e+00, 1.00000000e+00,
                        1.00000000e+00, 1.00000000e+00, 1.44422662e+00,
                        1.02998519e+00, 1.00000000e+00, 1.00000000e+00,
                        2.60322971e+01, 4.26534195e+01, 4.17046585e+01],
                       dtype='f').reshape(2, 3, 4, 5)
        self.assertTrue((vdfile.variables['UNKNOWN'] == checkv).all())


if __name__ == '__main__':
    unittest.main()



__all__ = ['temperature']
__doc__ = """
.. _Read
:mod:`Read` -- temperature Read interface
============================================

.. module:: Read
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` random access read for CAMx
              temperature files.  See PseudoNetCDF.sci_var.PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""

# Distribution packages
import unittest
import struct

# Site-Packages
from numpy import zeros, array, memmap

# This Package modules
from PseudoNetCDF.camxfiles.timetuple import timediff, timeadd, timerange
from PseudoNetCDF.camxfiles.FortranFileUtil import OpenRecordFile
from PseudoNetCDF.sci_var import PseudoNetCDFFile, PseudoNetCDFVariables


# for use in identifying uncaught nan
listnan = struct.unpack('>f', b'\xff\xc0\x00\x00')[0]
checkarray = zeros((1,), 'f')
checkarray[0] = listnan
array_nan = checkarray[0]


class temperature(PseudoNetCDFFile):
    """
    temperature provides a PseudoNetCDF interface for CAMx
    temperature files.  Where possible, the inteface follows
    IOAPI conventions (see www.baronams.com).

    ex:
        >>> temperature_path = 'camx_temperature.bin'
        >>> rows,cols = 65,83
        >>> temperaturefile = temperature(temperature_path,rows,cols)
        >>> temperaturefile.variables.keys()
        ['TFLAG', 'AIRTEMP', 'SURFTEMP']
        >>> tflag = temperaturefile.variables['TFLAG']
        >>> tflag.dimensions
        ('TSTEP', 'VAR', 'DATE-TIME')
        >>> tflag[0,0,:]
        array([2005185,       0])
        >>> tflag[-1,0,:]
        array([2005185,  240000])
        >>> v = temperaturefile.variables['SURFTEMP']
        >>> v.dimensions
        ('TSTEP', 'ROW', 'COL')
        >>> v.shape
        (25, 65, 83)
        >>> v = temperaturefile.variables['AIRTEMP']
        >>> v.dimensions
        ('TSTEP', 'LAY', 'ROW', 'COL')
        >>> v.shape
        (25, 28, 65, 83)
        >>> temperaturefile.dimensions
        {'TSTEP': 25, 'LAY': 28, 'ROW': 65, 'COL': 83}
    """

    id_fmt = 'fi'
    data_fmt = 'f'

    def __init__(self, rf, rows=None, cols=None):
        self.rffile = OpenRecordFile(rf)
        self.id_size = struct.calcsize(self.id_fmt)
        self.__readheader()
        self.__gettimestep()
        if rows is None and cols is None:
            rows = self.cell_count
            cols = 1
        elif rows is None:
            rows = self.cell_count / cols
        elif cols is None:
            cols = self.cell_count / rows
        else:
            if cols * rows != self.cell_count:
                raise ValueError(("The product of cols (%d) and rows (%d) " +
                                  "must equal cells (%d)") %
                                 (cols, rows, self.cell_count))

        self.createDimension('TSTEP', self.time_step_count)
        self.createDimension('COL', cols)
        self.createDimension('ROW', rows)
        self.createDimension('LAY', self.nlayers)
        self.createDimension('SURF', 1)
        self.variables = PseudoNetCDFVariables(
            self.__var_get, ['AIRTEMP', 'SURFTEMP'])

    def __var_get(self, key):
        def decor(k):
            return dict(units='K', var_desc=k.ljust(16),
                        long_name=k.ljust(16))

        def constr(k):
            return self.__variables(k)
        values = constr(key)
        dims = {'AIRTEMP': ('TSTEP', 'LAY', 'ROW', 'COL'),
                'SURFTEMP': ('TSTEP', 'SURF', 'ROW', 'COL')}[key]
        var = self.createVariable(key, 'f', dims)
        var[:] = values
        for k, v in decor(key).items():
            setattr(var, k, v)
        return var

    def __readheader(self):
        self.data_start_byte = 0
        self.rffile._newrecord(0)

        self.area_size = self.rffile.record_size
        self.area_count = (
            self.area_size - self.id_size) // struct.calcsize(self.data_fmt)
        self.area_padded_size = self.area_size + 8
        self.area_fmt = self.id_fmt + self.data_fmt * (self.area_count)

        self.start_time, self.start_date = self.rffile.read(self.id_fmt)

        self.record_size = self.rffile.record_size
        self.padded_size = self.record_size + 8
        self.cell_count = (
            self.record_size - self.id_size) // struct.calcsize(self.data_fmt)

        self.record_fmt = self.id_fmt + self.data_fmt * (self.cell_count)

    def __gettimestep(self):
        d, t = date, time = self.start_date, self.start_time
        self.nlayers = -1
        while (d, t) == (date, time):
            self.nlayers += 1
            t, d = self.rffile.read(self.id_fmt)
        self.time_step = timediff((self.start_date, self.start_time), (d, t))
        self.rffile.infile.seek(0, 2)
        self.rffile.previous()
        self.end_time, self.end_date = self.rffile.read(self.id_fmt)
        self.time_step_count = int(timediff(
            (self.start_date, self.start_time),
            (self.end_date, self.end_time)) // self.time_step) + 1

    def __variables(self, k):
        if k == 'SURFTEMP':
            out = zeros((len(self.dimensions['TSTEP']), 1, len(
                self.dimensions['ROW']), len(self.dimensions['COL'])), 'f')
            vars = self.__surfmaps()
        elif k == 'AIRTEMP':
            out = zeros((len(self.dimensions['TSTEP']),
                         len(self.dimensions['LAY']),
                         len(self.dimensions['ROW']),
                         len(self.dimensions['COL'])), 'f')
            vars = self.__airmaps()
        for i, v in enumerate(vars):
            out[i, ...] = v
        return out

    def __surfpos(self):
        pos = self.data_start_byte + 12
        inc = self.area_padded_size + self.padded_size * self.nlayers
        self.rffile.infile.seek(0, 2)
        rflen = self.rffile.tell()
        while pos < rflen:
            yield pos
            pos += inc

    def __surfmaps(self):
        for pos in self.__surfpos():
            tmpmm = memmap(self.rffile.infile.name, '>f', 'r', pos,
                           (self.area_count,))
            newshape = [len(self.dimensions['ROW']),
                        len(self.dimensions['COL'])]
            yield tmpmm.reshape(*newshape)

    def __airpos(self):
        pos = self.area_padded_size + self.data_start_byte
        inc = self.area_padded_size + self.padded_size * self.nlayers
        self.rffile.infile.seek(0, 2)
        rflen = self.rffile.tell()
        while pos < rflen:
            yield pos
            pos += inc

    def __airmaps(self):
        for pos in self.__airpos():
            firstshape = ((self.cell_count + 4) * self.nlayers,)
            tmpmm = memmap(self.rffile.infile.name, '>f', 'r', pos, firstshape)
            newshape1 = [self.nlayers, self.cell_count + 4]
            tmpmm = tmpmm.reshape(*newshape1)[:, 3:-1]
            newshape2 = [len(self.dimensions['LAY']),
                         len(self.dimensions['ROW']),
                         len(self.dimensions['COL'])]
            yield tmpmm.reshape(*newshape2)

    def timerange(self):
        return timerange((self.start_date, self.start_time),
                         timeadd((self.end_date, self.end_time),
                                 (0, self.time_step),
                                 (2400, 24)[int(self.time_step % 2)]),
                         self.time_step, (2400, 24)[int(self.time_step % 2)])


class TestRead(unittest.TestCase):
    def runTest(self):
        pass

    def setUp(self):
        pass

    def testTEMP(self):
        import PseudoNetCDF.testcase
        tpath = PseudoNetCDF.testcase.camxfiles_paths['temperature']
        tempfile = temperature(tpath, 4, 5)
        check_at = array([2.97762360e+02, 2.97261993e+02,
                          3.00761200e+02, 3.03811005e+02, 3.04561218e+02,
                          2.96350311e+02, 2.96676544e+02, 3.00992096e+02,
                          3.05474762e+02, 3.07840637e+02, 2.99522430e+02,
                          3.00271698e+02, 3.03738403e+02, 3.07201843e+02,
                          3.08288422e+02, 3.02957214e+02, 3.04927643e+02,
                          3.06630157e+02, 3.07726074e+02, 3.07380707e+02,
                          2.97516449e+02, 2.96920105e+02, 3.00340576e+02,
                          3.03413177e+02, 3.04202728e+02, 2.96074036e+02,
                          2.96250641e+02, 3.00632294e+02, 3.05113647e+02,
                          3.07390533e+02, 2.99310059e+02, 2.99901031e+02,
                          3.03344666e+02, 3.06782135e+02, 3.07819946e+02,
                          3.02657013e+02, 3.04522675e+02, 3.06167206e+02,
                          3.07235107e+02, 3.06883484e+02, 2.97677338e+02,
                          2.96919098e+02, 3.00031250e+02, 3.03082672e+02,
                          3.03850861e+02, 2.96460999e+02, 2.95947815e+02,
                          3.00303680e+02, 3.04781982e+02, 3.07048492e+02,
                          2.99246979e+02, 2.99508667e+02, 3.02997650e+02,
                          3.06450500e+02, 3.07478485e+02, 3.02246765e+02,
                          3.04192139e+02, 3.05832489e+02, 3.06897644e+02,
                          3.06546173e+02, 2.97428253e+02, 2.97174896e+02,
                          3.00208191e+02, 3.03096893e+02, 3.04174133e+02,
                          2.96558685e+02, 2.96706177e+02, 3.00862610e+02,
                          3.04807037e+02, 3.06937347e+02, 2.98850220e+02,
                          2.99482727e+02, 3.03085022e+02, 3.06456787e+02,
                          3.07406586e+02, 3.01888580e+02, 3.03996735e+02,
                          3.05916962e+02, 3.07113647e+02, 3.06539337e+02,
                          2.97645966e+02, 2.97326630e+02, 3.00117950e+02,
                          3.02804077e+02, 3.03801544e+02, 2.96783661e+02,
                          2.96694946e+02, 3.00722931e+02, 3.04501587e+02,
                          3.06560150e+02, 2.98854828e+02, 2.99314972e+02,
                          3.02861023e+02, 3.06150177e+02, 3.07073944e+02,
                          3.01700745e+02, 3.03746124e+02, 3.05626617e+02,
                          3.06770447e+02, 3.06172394e+02, 2.97927094e+02,
                          2.97691681e+02, 3.00104675e+02, 3.02464874e+02,
                          3.03398926e+02, 2.97336578e+02, 2.97074127e+02,
                          3.00716736e+02, 3.04132446e+02, 3.06129700e+02,
                          2.98817017e+02, 2.99221039e+02, 3.02649536e+02,
                          3.05787415e+02, 3.06698334e+02, 3.01333618e+02,
                          3.03411346e+02, 3.05317505e+02, 3.06446869e+02,
                          3.05815948e+02], dtype='f').reshape(2, 3, 4, 5)
        self.assertTrue((tempfile.variables['AIRTEMP'] == check_at).all())


if __name__ == '__main__':
    unittest.main()



__all__ = ['irr']
__doc__ = """
.. _Read
:mod:`Read` -- irr Read interface
============================================

.. module:: Read
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` random access read for CAMx
              irr files.  See PseudoNetCDF.sci_var.PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""

# Distribution packages
from datetime import datetime
import unittest
import struct
from warnings import warn
from math import ceil

# Site-Packages
from numpy import zeros, array, dtype, fromfile, arange

# This Package modules
from PseudoNetCDF.sci_var import PseudoNetCDFFile, PseudoNetCDFVariable
from PseudoNetCDF.sci_var import PseudoNetCDFVariables
pncvar = PseudoNetCDFVariable

# for use in identifying uncaught nan
listnan = struct.unpack('>f', b'\xff\xc0\x00\x00')[0]
checkarray = zeros((1,), 'f')
checkarray[0] = listnan
array_nan = checkarray[0]


class irr(PseudoNetCDFFile):
    """
    irr provides a PseudoNetCDF interface for CAMx
    irr files.  Where possible, the inteface follows
    IOAPI conventions (see www.baronams.com).

    ex:
        >>> irr_path = 'camx_irr.bin'
        >>> irrfile = irr(irr_path)
        >>> irrfile.variables.keys()
        ['TFLAG', 'RXN_01', 'RXN_02', 'RXN_03', ...]
        >>> v = irrfile.variables['RXN_01']
        >>> tflag = irrfile.variables['TFLAG']
        >>> tflag.dimensions
        ('TSTEP', 'VAR', 'DATE-TIME')
        >>> tflag[0,0,:]
        array([2005185,       0])
        >>> tflag[-1,0,:]
        array([2005185,  240000])
        >>> v.dimensions
        ('TSTEP', 'LAY', 'ROW', 'COL')
        >>> v.shape
        (25, 28, 65, 83)
        >>> irrfile.dimensions
        {'TSTEP': 25, 'LAY': 28, 'ROW': 65, 'COL': 83}
    """

    id_fmt = "ifiiiii"
    data_fmt = "f"

    def __init__(self, rf, units='ppm/hr', conva=None, nvarcache=None):
        """
        Initialization included reading the header and learning
        about the format.

        see __readheader and __gettimestep() for more info
        """
        rffile = self._rffile = open(rf)
        rffile.seek(0, 2)
        if rffile.tell() < 2147483648:
            warn("For greater speed on files <2GB use ipr_memmap")
        rffile.seek(0, 0)
        self.__readheader()
        self.__gettimestep()
        self.units = units
        # __conv is a conversion array that comes from ipr
        # if units is not ppm/hr, conv must be provided
        self.__conv = conva
        if self.units != 'ppm/hr' and self.__conv is None:
            raise ValueError("When units are provided, a conversion array " +
                             "dim(t,z,x,y) must also be provided")
        varkeys = ['RXN_%02d' % i for i in range(1, self.NRXNS + 1)]

        domain = self.padomains[0]
        tdim = self.createDimension('TSTEP', int(self.NSTEPS))
        tdim.setunlimited(True)
        self.createDimension('COL', int(domain['iend'] - domain['istart'] + 1))
        self.createDimension('ROW', int(domain['jend'] - domain['jstart'] + 1))
        self.createDimension('LAY', int(domain['tlay'] - domain['blay'] + 1))
        self.createDimension('DATE-TIME', 2)
        self.createDimension('VAR', int(self.NRXNS))
        self.variables = PseudoNetCDFVariables(self.__var_get, varkeys)
        self._nvarcache = nvarcache or int(ceil(self.NRXNS / 4.))
        tflag = self.createVariable(
            'TFLAG', 'i', ('TSTEP', 'VAR', 'DATE-TIME'))
        tflag.units = '<YYYYJJJ, HHMMSS>'
        tflag.var_desc = tflag.long_name = 'TFLAG'.ljust(16)
        tflag[:] = array(self.timerange(), dtype='i').reshape(
            self.NSTEPS, 1, 2)

    def __var_get(self, key):
        rxni = int(key.split('_')[1])
        self.loadVars(rxni, self._nvarcache)
        tflag = self.createVariable(
            'TFLAG', 'i', ('TSTEP', 'VAR', 'DATE-TIME'))
        tflag.units = '<YYYYJJJ, HHMMSS>'
        tflag.var_desc = tflag.long_name = 'TFLAG'.ljust(16)
        tflag[:] = array(self.timerange(), dtype='i').reshape(
            self.NSTEPS, 1, 2)
        return self.variables[key]

    def __readheader(self):
        """
        __readheader reads the header section of the ipr file
        it initializes each header field (see CAMx Users Manual for a list)
        as properties of the ipr class
        """
        rffile = self._rffile
        notedt = dtype([('SPAD', '>i'), ('NOTE', 'S80'), ('EPAD', '>i')])
        self.runmessage = fromfile(rffile, dtype=notedt, count=1)[0]['NOTE']
        datedt = dtype([('SPAD', '>i'), ('SDATE', '>i'), ('STIME', '>f'),
                        ('EDATE', '>i'), ('ETIME', '>f'), ('EPAD', '>i')])
        self.SDATE, self.STIME, self.EDATE, self.ETIME = \
            fromfile(rffile, dtype=datedt, count=1)[0].tolist()[1:-1]

        self.grids = []
        ngdt = dtype([('SPAD', '>i'), ('NGRIDS', '>i'), ('EPAD', '>i')])
        gdefdt = dtype([('SPAD', '>i'), ('orgx', '>f'), ('orgy', '>f'),
                        ('ncol', '>i'), ('nrow', '>i'), ('xsize', '>f'),
                        ('ysize', '>f'), ('iutm', '>i'), ('EPAD', '>i')])
        for grid in range(fromfile(rffile, dtype=ngdt, count=1)[0]['NGRIDS']):
            gdefval = fromfile(rffile, dtype=gdefdt, count=1)[0].tolist()[1:-1]
            self.grids.append(
                dict(
                    zip(
                        gdefdt.names[1:-1],
                        gdefval
                    )
                )
            )

        self.padomains = []
        paddt = dtype([('SPAD', '>i'), ('NPADOMAINS', '>i'), ('EPAD', '>i')])
        padefdt = dtype([('SPAD', '>i'), ('grid', '>i'), ('istart', '>i'),
                         ('iend', '>i'), ('jstart', '>i'), ('jend', '>i'),
                         ('blay', '>i'), ('tlay', '>i'), ('EPAD', '>i')])
        padn = fromfile(rffile, dtype=paddt, count=1)[0]['NPADOMAINS']
        for padomain in range(padn):
            padef = fromfile(rffile, dtype=padefdt, count=1)[0].tolist()[1:-1]
            self.padomains.append(
                dict(
                    zip(
                        padefdt.names[1:-1],
                        padef
                    )
                )
            )
        nrxndt = dtype([('SPAD', '>i'), ('NRXNS', '>i'), ('EPAD', '>i')])
        self.NRXNS = fromfile(rffile, dtype=nrxndt, count=1)[0]['NRXNS']

        self._data_start_byte = rffile.tell()

        self._record_dtype = dtype([('SPAD', '>i'), ('DATE', '>i'),
                                    ('TIME', '>f'), ('PAGRID', '>i'),
                                    ('NEST', '>i'), ('I', '>i'), ('J', '>i'),
                                    ('K', '>i'), ('IRRS', '>f', self.NRXNS),
                                    ('EPAD', '>i')])
        self._padded_size = self._record_dtype.itemsize

    def __gettimestep(self):
        """
        Header information provides start and end date, but does not
        indicate the increment between.  This routine reads the first
        and second date/time and initializes variables indicating the
        timestep length and the anticipated number.
        """
        self._activedomain = self.padomains[0]
        self._rffile.seek(
            self._data_start_byte + (
                self.__jrecords(0, self.padomains[0]['jend']) *
                self._padded_size
            ), 0
        )
        date, time = fromfile(self._rffile, dtype=self._record_dtype, count=1)[
            0].tolist()[1:3]
        tstart = datetime.strptime('%05dT%04d' % (
            self.SDATE, self.STIME), '%y%jT%H%M')
        tone = datetime.strptime('%05dT%04d' % (date, time), '%y%jT%H%M')
        tstep = tone - tstart
        self.time_step = self.TSTEP = int(
            (datetime.strptime('0000', '%H%M') + tstep).strftime('%H%M'))
        tend = datetime.strptime('%05dT%04d' % (
            self.EDATE, self.ETIME), '%y%jT%H%M')
        tdiff = tend - tstart
        multiple = (tdiff.days * 24 * 3600. + tdiff.seconds) / \
            (tstep.days * 24 * 3600. + tstep.seconds)
        self.NSTEPS = self.time_step_count = int(multiple)
        assert(multiple == int(multiple))

    def __gridrecords(self, pagrid):
        """
        routine returns the number of records to increment from the
        data start byte to find the pagrid
        """
        ntime = self.__timerecords(
            pagrid, (self.EDATE, int(self.ETIME + self.TSTEP)))
        return ntime

    def __timerecords(self, pagrid, dt):
        """
        routine returns the number of records to increment from the
        data start byte to find the first time
        """
        d, t = dt
        nsteps = self.timerange().index((d, t))
        nj = self.__jrecords(pagrid, self.padomains[pagrid]['jend'] + 1)
        return nsteps * nj

    def __irecords(self, pagrid, i):
        """
        routine returns the number of records to increment from the
        data start byte to find the first icell
        """
        ni = i - self._activedomain['istart']
        nk = self.__krecords(pagrid, self._activedomain['tlay'] + 1)
        return ni * nk

    def __jrecords(self, pagrid, j):
        """
        routine returns the number of records to increment from the
        data start byte to find the first jcell
        """
        nj = j - self._activedomain['jstart']
        ni = self.__irecords(pagrid, self._activedomain['iend'] + 1)
        return nj * ni

    def __krecords(self, pagrid, k):
        """
        routine returns the number of records to increment from the
        data start byte to find the first kcell
        """
        return k - self._activedomain['blay']

    def __recordposition(self, pagrid, date, time, i, j, k):
        """
        routine uses pagridrecords, timerecords,irecords,
        jrecords, and krecords multiplied by the fortran padded size
        to return the byte position of the specified record

        pagrid - integer
        date - integer
        time - float
        i - integer
        j - integer
        k - integer
        """
        records = 0
        for pag in range(pagrid):
            records += self.__gridrecords(pag)
        records += self.__timerecords(pagrid, (date, time))
        records += self.__jrecords(pagrid, j)
        records += self.__irecords(pagrid, i)
        records += self.__krecords(pagrid, k)
        return records * self._padded_size + self._data_start_byte

    def _seek(self, pagrid=1, date=None, time=None, i=1, j=1, k=1):
        """
        Move file cursor to beginning of specified record
        see __recordposition for a definition of variables
        """
        if date is None:
            date = self.SDATE
        if time is None:
            time = self.STIME + self.TSTEP
        self._activedomain = self.padomains[pagrid]
        self._rffile.seek(self.__recordposition(pagrid, date, time, i, j, k))

    def loadVars(self, start, n, pagrid=0):
        domain = self.padomains[pagrid]
        istart = domain['istart']
        iend = domain['iend']
        jstart = domain['jstart']
        jend = domain['jend']
        kstart = domain['blay']
        kend = domain['tlay']
        variables = self.variables
        nk = kend + 1 - kstart
        nj = jend + 1 - jstart
        ni = iend + 1 - istart
        nrec = nk * ni * nj
        temp = zeros((nrec, self.NRXNS), 'f')
        shape = (self.NSTEPS,) + \
            tuple(eval('map(len, (LAY, ROW, COL))', None, self.dimensions))
        variables.clear()
        end = min(start + n, self.NRXNS + 1)
        start = max(1, end - n)
        for rxn in range(start, end):
            key = 'RXN_%02d' % rxn
            variables[key] = pncvar(self, key, 'f',
                                    ('TSTEP', 'LAY', 'ROW', 'COL'),
                                    values=zeros(shape, 'f'), units='ppm/hr',
                                    var_desk=key.ljust(16),
                                    long_name=key.ljust(16))

        self._seek(pagrid=0, i=istart, j=jstart, k=kstart)
        for ti, (d, t) in enumerate(self.timerange()):
            record = fromfile(
                self._rffile, dtype=self._record_dtype, count=nrec)
            temp[:] = record['IRRS']
            date = record['DATE']
            time = record['TIME']
            id = record['I']
            jd = record['J']
            kd = record['K']
            assert((id == arange(istart, iend + 1)
                    [None, :, None].repeat(nk, 2).repeat(nj, 0).ravel()).all())
            assert((kd == arange(kstart, kend + 1)
                    [None, None, :].repeat(ni, 1).repeat(nj, 0).ravel()).all())
            assert(((jd == arange(jstart, jend + 1).repeat(ni * nk, 0))).all())
            assert((date == d).all() and (time == t).all())
            for rxn in range(start, end):
                variables['RXN_%02d' % rxn][ti, :, :, :] = \
                    temp[:, rxn - 1].reshape(nj, ni, nk).swapaxes(1, 2)\
                                    .swapaxes(0, 1)

    def timerange(self):
        tstart = datetime.strptime('%05dT%04d' % (
            self.SDATE, self.STIME), '%y%jT%H%M')
        tdiff = datetime.strptime(
            '%04d' % self.TSTEP, '%H%M') - datetime.strptime('0000', '%H%M')
        dates = [tstart + (tdiff * i) for i in range(1, self.NSTEPS + 1)]
        return [(int(d.strftime('%y%j')), float(d.strftime('%H%M')))
                for d in dates]


class TestRead(unittest.TestCase):
    def runTest(self):
        pass

    def setUp(self):
        pass

    def testIRR(self):
        warn('Test not implemented; data too big for distribution')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
from __future__ import print_function
import numpy as np
from warnings import warn
from netCDF4 import Dataset
from collections import defaultdict
from datetime import datetime, timedelta
from glob import glob

from PseudoNetCDF.pncparse import AggCommaString

unitconvert = {('ppmV', 'ppb'): lambda x: x * 1000.}


def add_vertprofile_options(vertparser):
    vpadd = vertparser.add_argument
    vpadd("--sigma", dest="sigma", action="store_true", default=False,
          help="Plot on sigma coordinate instead of pressure")

    vpadd("--scale", dest="scale", type=str, default='log',
          help="Defaults to log, but linear and semilog are also options.")

    vpadd("--minmax", dest="minmax", type=str, default="None,None",
          help="Use values to set range (xmin, xmax); defaults None,None.")

    vpadd("--mask-zeros", dest="maskzeros", action="store_true",
          default=False, help="Defaults False.")

    vpadd("--minmaxq", dest="minmaxq", type=str, default='0,100',
          help="Use quartiles to set range (xmin, xmax); defaults 0,100.")

    vpadd("--out-unit", dest="outunit", type=str, default='ppb',
          help="Defaults ppb.")

    vpadd("--tes-paths", dest="tespaths", type=str,
          action=AggCommaString, default='',
          help="Plot tes on top of boundary from paths; defaults to []")

    vpadd("--omi-paths", dest="omipaths", type=str,
          default='', action=AggCommaString,
          help="Plot omi on top of boundary from paths; defaults to []")

    vpadd("--itertime", dest="itertime", default=False, action='store_true',
          help="Iterate over times and plot each one.")

    vpadd("--edges", dest="edges", default=False, action="store_true",
          help="Plot S,E,N,W edges instead of a single plot.")

# vertparser = getparser(has_ofile = True, plot_options = True,
#                        interactive = False)
# add_vertprofile_options(vertparser)
# vertparser.epilog = """
# Example:
# $ pncvertprofile.py outputs/ts20120301.bpch.BCON.nc test_profile -v O3 \
#    --edges #--minmaxq .5,99.5 --tes-paths=~/Data/test/*
# """


def maskfilled(v):
    return np.ma.masked_values(v[:], v.attrs['_FillValue'])


def plot_omi(ax, lon_bnds, lat_bnds, omipaths, key='O3Profile',
             airden=None, airdenvert=None):
    import h5py
    # from scipy.constants import Avogadro
    allx = []
    ally = []
    lon_bnds = lon_bnds[:]
    lat_bnds = lat_bnds[:]
    if len(omipaths) == 0:
        return None, None
    outomipaths = []
    for omipath in omipaths:
        outomipaths.extend(glob(omipath))
    omipaths = outomipaths
    omipaths.sort()
    for path in omipaths:
        print(path)
        f = h5py.File(path, mode='r')
        swaths = f['HDFEOS']['SWATHS'][key]
        latsv = swaths['Geolocation Fields']['Latitude']
        lonsv = swaths['Geolocation Fields']['Longitude']
        lats = maskfilled(latsv).reshape(-1, 1)
        lons = maskfilled(lonsv).reshape(-1, 1)
        inboth = matchspace(lons, lats, lon_bnds, lat_bnds)
        if inboth.filled(False).sum(0) > 0:
            print('******** FOUND ******', path)
            pressurev = swaths['Geolocation Fields']['Pressure']
            altitudev = swaths['Geolocation Fields']['Altitude']
            nk = pressurev.shape[-1]
            speciesv = swaths['Data Fields']['O3']
            species = maskfilled(speciesv).reshape(-1, nk - 1)

            # Pressure is from top to bottom
            pressure = maskfilled(pressurev).reshape(-1, nk)
            altitude = maskfilled(altitudev).reshape(-1, nk)
            dz = -(altitude[:, 1:] - altitude[:, :-1]) * 1000. * 100.
            # Lacis 1990 1 DU = 2.69e16 molecules cm-2
            species = species * .001 / dz  # cm /cm # vrm
            x = pressure[inboth]
            y = species[inboth]
            allx.append(x)
            ally.append(y)
        else:
            wtmp = 'No data found for %s: lons (%s, %s) and lats (%s, %s); '
            wtmp += 'lonbs (%s, %s) and latbs (%s, %s);'
            wvals = (path, lons.min(), lons.max(), lats.min(), lats.max())
            wvals = wvals + (lon_bnds.min(), lon_bnds.max(), lat_bnds.min())
            wvals = wvals + (lat_bnds.max(),)
            warn(wtmp % wvals)

    if len(allx) == 0:
        print('*' * 80 + '\n\nNo OMI DATA FOUND AT ALL\n\n' + '*' * 80)
        return None, None

    var = np.ma.masked_values(np.ma.concatenate(ally, axis=0), -999.) * 1e9
    var = var.reshape(-1, var.shape[-1])
    vertcrd = np.ma.masked_values(np.ma.concatenate(
        allx, axis=0), -999.).mean(0)  # .reshape(-1, 2).mean(1)
    omil, omir = minmaxmean(ax, var.T.repeat(2, 0), vertcrd.repeat(2, 0)[1:-1],
                            ls='-', lw=2, color='b', facecolor='b',
                            edgecolor='b', alpha=.2, zorder=5, label='OMI')
    ax.text(.05, .7, 'OMI = %d' % var.shape[0], transform=ax.transAxes)
    return omil, omir


def matchspace(lons, lats, lon_bnds, lat_bnds):
    slicer = [slice(None, None)] + [None] * (lon_bnds.ndim - 1)
    inlon = np.logical_and(lons[slicer] >= lon_bnds[None, ..., 0],
                           lons[slicer] <= lon_bnds[None, ..., 1])
    inlat = np.logical_and(lats[slicer] >= lat_bnds[None, ..., 0],
                           lats[slicer] <= lat_bnds[None, ..., 1])
    inboth = np.logical_and(inlon, inlat).reshape(inlon.shape[0], -1).any(1)
    return inboth


def plot_tes(ax, lon_bnds, lat_bnds, tespaths):
    allx = []
    ally = []
    lon_bnds = lon_bnds[:]
    lat_bnds = lat_bnds[:]
    if len(tespaths) == 0:
        return None, None
    outtespaths = []
    for tespath in tespaths:
        outtespaths.extend(glob(tespath))
    tespaths = outtespaths
    tespaths.sort()
    for path in tespaths:
        f = Dataset(path)
        lats = f.variables['latitude'][:][:, None]
        lons = f.variables['longitude'][:][:, None]
        pressure = f.variables['pressure']
        species = f.variables['species']
        inboth = matchspace(lons, lats, lon_bnds, lat_bnds)
        if inboth.sum(0) > 0:
            print('******** FOUND ******', path)
            x = pressure[inboth]
            y = species[inboth]
            allx.append(x)
            ally.append(y)
        else:
            warn('No data found for %s' % path)

    if len(allx) == 0:
        return None, None
    var = np.ma.masked_values(np.ma.concatenate(ally, axis=0), -999.) * 1e9
    var = var.reshape(-1, var.shape[-1])
    vertcrd = np.ma.masked_values(
        np.ma.concatenate(allx, axis=0), -999.).mean(0)
    tesl, tesr = minmaxmean(ax, var.T, vertcrd, ls='-', lw=2, color='r',
                            facecolor='r', edgecolor='r', alpha=.2, zorder=2,
                            label='TES (%d)' % var.shape[0])
    return tesl, tesr


def minmaxmean(ax, vals, vertcrd, **kwds):
    minval = vals.min(1)
    meanval = vals.mean(1)
    maxval = vals.max(1)
    linekwds = kwds.copy()
    linekwds['color'] = linekwds.pop('facecolor')
    linekwds.pop('edgecolor')
    linekwds.pop('alpha')
    fillkwds = kwds.copy()
    fillkwds['ls'] = 'solid'

    line, = ax.plot(meanval, vertcrd, **linekwds)

    x = np.ma.concatenate([minval[:vertcrd.size], maxval[:vertcrd.size][::-1]])
    y = np.ma.concatenate([vertcrd[:], vertcrd[::-1]])
    mask = x.mask | y.mask
    x = np.ma.masked_where(mask, x).compressed()
    y = np.ma.masked_where(mask, y).compressed()
    range, = ax.fill(x, y, **fillkwds)
    return line, range


def plotprofile(args):
    vertprofileplot(args.ifiles, args)


def vertprofileplot(ifiles, args):
    if args.variables is None:
        raise ValueError('User must specify variable(s) to plot:\n%s' %
                         '\n\t'.join(ifiles[0].variables.keys()))
    from PseudoNetCDF.coordutil import getsigmamid, getpresmid, gettimes
    import matplotlib.pyplot as plt
    plt.rcParams['text.usetex'] = False
    # from matplotlib.colors import LinearSegmentedColormap, BoundaryNorm
    # from matplotlib.colors import LogNorm
    scale = args.scale
    minmax = eval(args.minmax)
    minmaxq = eval(args.minmaxq)
    sigma = args.sigma
    maskzeros = args.maskzeros
    outunit = args.outunit
    tespaths = args.tespaths
    omipaths = args.omipaths
    edges = args.edges
    try:
        f, = ifiles
    except Exception:
        raise ValueError('curtain plot expects one file when done. ' +
                         'Try stack time --stack=time to concatenate')

    # Add CF conventions if necessary
    if 'latitude_bounds' not in f.variables.keys():
        try:
            from PseudoNetCDF import getvarpnc
            from PseudoNetCDF.conventions.ioapi import add_cf_from_ioapi
            f = getvarpnc(f, None)
            add_cf_from_ioapi(f)
        except Exception:
            pass
    if sigma:
        vertcrd = getsigmamid(f)
    else:
        vertcrd = getpresmid(f, pref=101325., ptop=getattr(f, 'VGTOP', 10000))
        if vertcrd.max() > 2000:
            vertcrd /= 100.

    try:
        lonb = f.variables['geos_longitude_bounds']
        latb = f.variables['geos_latitude_bounds']
    except Exception:
        lonb = f.variables['longitude_bounds']
        latb = f.variables['latitude_bounds']
    for var_name in args.variables:
        temp = defaultdict(lambda: 1)
        try:
            eval(var_name, None, temp)
            var = eval(var_name, None, f.variables)[:]
        except Exception:
            temp[var_name]
            var = f.variables[var_name][:]
        if maskzeros:
            var = np.ma.masked_values(var, 0)
        vkeys = [k for k in temp.keys()]
        unit = f.variables[vkeys[0]].units.strip()
        if unit in unitconvert:
            var = unitconvert.get((unit, outunit), lambda x: x)(var)
        else:
            outunit = unit
        vmin, vmax = np.percentile(
            np.ma.compressed(var).ravel(), list(minmaxq))
        if minmax[0] is not None:
            vmin = minmax[0]
        if minmax[1] is not None:
            vmax = minmax[1]
        if edges:
            fig = plt.figure(figsize=(16, 4))
            offset = 0.05
            ax = fig.add_axes([.1 - offset, .15, .22, .725])
            ax = fig.add_axes([.325 - offset, .15, .22, .725])
            ax = fig.add_axes([.55 - offset, .15, .22, .725])
            ax = fig.add_axes([.775 - offset, .15, .22, .725])
            ss = 0
            se = ss + f.NCOLS + 1
            es = se
            ee = se + f.NROWS + 1
            ns = ee
            ne = ee + f.NCOLS + 1
            ws = ne
            we = ws + f.NROWS + 1
            axs = fig.axes
            for ax in fig.axes[1:]:
                ax.yaxis.set_major_formatter(plt.NullFormatter())

            vars = [var[:, :, ss:se], var[:, :, es:ee], var[:, :, ns:ne]
                    [:, :, ::-1], var[:, :, ws:we][:, :, ::-1]]
            lonbss = [lonb[ss:se], lonb[es:ee],
                      lonb[ns:ne][::-1], lonb[ws:we][::-1]]
            latbss = [latb[ss:se], latb[es:ee],
                      latb[ns:ne][::-1], latb[ws:we][::-1]]

        else:
            fig = plt.figure()
            ax = fig.add_subplot(111)
            axs = fig.axes
            vars = [var]
            ismultidimcoord = (lonb.dimensions == ('longitude', 'nv') and
                               latb.dimensions == ('latitude', 'nv'))
            if ismultidimcoord:
                lonbss = [lonb[:][None, :, :]]
                latbss = [latb[:][:, None, :]]
            else:
                lonbss = [lonb[:]]
                latbss = [latb[:]]
        for ax, var, lonbs, latbs in zip(axs, vars, lonbss, latbss):
            vals = var.swapaxes(0, 1).reshape(var.shape[1], -1)
            modl, modr = minmaxmean(ax, vals, vertcrd, facecolor='k',
                                    edgecolor='k', alpha=.2, zorder=4,
                                    label='mod (%d)' % vals.shape[1],
                                    ls='-', lw=2, color='k')
            llines = [(modl, modr)]
            ymin, ymax = vertcrd.min(), vertcrd.max()
            ax.set_ylim(ymax, ymin)
            ax.set_xscale(scale)
            ax.set_xlim(vmin, vmax)
            # if scale == 'log':
            #   ax.set_xticklabels(['%.1f' % (10**x) for x in ax.get_xticks()])

            if 'TFLAG' in f.variables.keys():
                SDATE = f.variables['TFLAG'][:][0, 0, 0]
                EDATE = f.variables['TFLAG'][:][-1, 0, 0]
                STIME = f.variables['TFLAG'][:][0, 0, 1]
                ETIME = f.variables['TFLAG'][:][-1, 0, 1]
                if SDATE == 0:
                    SDATE = 1900001
                    EDATE = 1900001
                sdate = datetime.strptime(
                    '%07d %06d' % (SDATE, STIME), '%Y%j %H%M%S')
                edate = datetime.strptime(
                    '%07d %06d' % (EDATE, ETIME), '%Y%j %H%M%S')
            elif 'tau0' in f.variables.keys():
                sdate = datetime(1985, 1, 1, 0) + \
                    timedelta(hours=f.variables['tau0'][0])
                edate = datetime(1985, 1, 1, 0) + \
                    timedelta(hours=f.variables['tau1'][-1])
            else:
                times = gettimes(f)
                sdate = times[0]
                edate = times[-1]

            if len(tespaths) > 0:
                tesl, tesr = plot_tes(ax, lonbs, latbs, tespaths)
                if tesl is not None:
                    llines.append((tesl, tesr))
            if len(omipaths) > 0:
                airdenvals = f.variables['AIRDEN'][:].mean(0).mean(1)
                omil, omir = plot_omi(ax, lonbs, latbs, omipaths,
                                      airden=airdenvals, airdenvert=vertcrd)
                if omil is not None:
                    llines.append((omil, omir))

        try:
            title = '%s to %s' % (sdate.strftime(
                '%Y-%m-%d'), edate.strftime('%Y-%m-%d'))
        except Exception:
            title = var_name
        if sigma:
            axs[0].set_ylabel('sigma')
        else:
            axs[0].set_ylabel('pressure')

        xmax = -np.inf
        xmin = np.inf
        for ax in fig.axes:
            tmp_xmin, tmp_xmax = ax.get_xlim()
            xmax = max(tmp_xmax, xmax)
            xmin = min(tmp_xmin, xmin)
        for ax in fig.axes:
            ax.set_xlim(xmin, xmax)

        if len(axs) == 1:
            axs[0].set_xlabel('%s %s' % (var_name, outunit))
        else:
            axs[0].set_xlabel('South')
            axs[1].set_xlabel('East')
            axs[2].set_xlabel('North')
            axs[3].set_xlabel('West')
            fig.text(.5, .90, '%s %s' % (var_name, outunit),
                     horizontalalignment='center', fontsize=16)
        nl = 0
        for ax in axs:
            if len(ax.get_lines()) > nl:
                nl = len(ax.get_lines())
                plt.sca(ax)

        llabels = [l[0].get_label() for l in llines]
        plt.legend(llines, llabels, bbox_to_anchor=(.1, 1),
                   loc='upper left', bbox_transform=fig.transFigure, ncol=6)
        if edges:
            fig.text(0.95, 0.975, title, horizontalalignment='right',
                     verticalalignment="top", fontsize=16)
        else:
            fig.text(0.95, 0.025, title, horizontalalignment='right',
                     verticalalignment="bottom", fontsize=16)
        figpath = args.outpath + var_name + '.' + args.figformat
        fig.savefig(figpath)
        if args.verbose > 0:
            print('Saved fig', figpath)
        # plt.close(fig)
    return fig



#!/usr/bin/env python
import matplotlib.pyplot as plt
from PseudoNetCDF.coordutil import gettimes
import numpy as np


def plot_diurnal_box(ifiles, args):
    times = [gettimes(ifile) for ifile in ifiles]
    hours = [np.array([t.hour for t in time]) for time in times]
    fig = plt.figure()
    sax = fig.add_subplot(111)
    if len(args.figure_keywords) > 0:
        plt.setp(fig, **args.figure_keywords)
    if len(args.axes_keywords) > 0:
        plt.setp(sax, **args.axes_keywords)
    sax.set_xlabel('Time (UTC)')
    for target in args.variables:
        vars = [ifile.variables[target] for ifile in ifiles]
        unit = getattr(vars[0], 'units', 'unknown')
        sax.set_ylabel(target + ' (' + unit + ')')
        vals = [var[:] for var in vars]
        hvars = [[np.ma.compressed(val[hour == i]) for i in range(24)]
                 for hour, val in zip(hours, vals)]
        del sax.lines[:]
        nvars = len(vars)
        varwidth = .8 / nvars / 1.1
        po = np.arange(24) + 0.1 + varwidth / 2
        ncolors = max(float(nvars), 10)
        try:
            from cycler import cycler
            props = cycler('color', [plt.get_cmap()(
                (nc + .5) / float(ncolors)) for nc in range(ncolors)])
        except Exception:
            props = [dict(color=plt.get_cmap()((nc + .5) / float(ncolors)))
                     for nc in range(ncolors)]
        for vi, (var, propd) in enumerate(zip(hvars, props)):
            varb = sax.boxplot(var, positions=po + vi * varwidth *
                               1.1, widths=varwidth, patch_artist=True)
            plt.setp([i for i in varb.values()], **propd)
            plt.setp(varb['medians'], color='k')
            plt.setp(varb['fliers'], markeredgecolor=propd['color'])
        sax.set_xlim(-.5, 24.5)
        sax.set_xticks(range(0, 25))
        sax.set_xticklabels([str(i) for i in range(0, 25)])
        figpath = args.outpath + target + '.' + args.figformat
        fig.savefig(figpath)
        if args.verbose > 0:
            print('Saved fig', figpath)


if __name__ == '__main__':
    from PseudoNetCDF.pncparse import getparser, pncparse
    parser = getparser(plot_options=True, has_ofile=True)
    parser.add_argument('--whis', default=None,
                        help='See pydoc matplotlib.axes.Axes.boxplot')
    parser.epilog += """
    -----
box.py inobs inmod target [target ...]
inobs - path to obs
inmod - path to mod
target - variable name
"""
    ifiles, args = pncparse(plot_options=True, has_ofile=True, parser=parser)
    plot_diurnal_box(ifiles, args)



"""
NOAA/GMD sounding data file produced on 2015/02/02 22:04:08 GMT ...
               Header lines = 45
               Data columns = 21
              Flight number = BL049
                 Date [GMT] = 03-01-1992
                 Time [GMT] = 18:35:21
                   Location = Boulder, CO
                  Longitude = -105.25000
                   Latitude = 40.00000
       Launch altitude (km) = 1.743
      Surface pressure (mb) = 826.790
  Surface temperature (deg) = 6.870
       Surface humidity (%) = 18.900
         Turn altitude (km) = 30.699
         Turn pressure (mb) = 10.20
          Radiosonde number = 8815949
     Vaisala humicap sensor = A
       Radiation correction = Yes
     Pressure sensor offset = 0.000
                 A/D System = 12 bit Tmax
Radiosonde Total Col. Water = 2.322
            Instrument Type = Ozone Sonde
       Original File Source = unknown
         Ozone sonde number = 5A8087
                   Solution = 1%
Oltmans solution correction = Yes
             Oltmans term A = 1.0000
             Oltmans term B = 0.4000
   Total ozone column (CMR) = 283 (41)
  Total ozone column (SBUV) = 289 (47)
  Total ozone stop pressure = 10.28
  Time (sec) to pump 100 ml = 28.230
    Dry flowrate correction = 2.60
    Background current (uA) = 0.071
               Coefficients =
       Pump coefficient pc0 = 0.5955
                        pc1 = 0.5125
                        pc2 = -0.2353
                        pc3 = 0.0824
       Pressure data source = Radiosonde Pressure
       Altitude data source = Radiosonde Geopot Altitude
   Geo alt anchoring method = First/Launch Row Set to Launch Altitude

      Time,     Press,      Praw,       Alt,     Tcorr,      Temp,..."""

# Time, Press, Praw, Alt, Tcorr, Temp, Traw, Theta, RH, RHraw, TFp V,
# IPW V, TVaisI, O3 Cell I, O3 P, O3 Mr, T Pump, O3Bat, I Pump,
# Total Column O3, Total w/ Extrap O3


from PseudoNetCDF import PseudoNetCDFFile
import re
import numpy as np
spaces = re.compile(r',\s{0,1000}')


def skysonde1sec(inpath):
    datafile = open(inpath, 'r')
    datalines = datafile.read().split('\n')
    nmeta = int(datalines[1].split(' = ')[1])
    meta = dict([[w.strip() for w in line.split(' = ')]
                 for line in datalines[1:nmeta - 2] if line != ''])

    varline, unitline = datalines[nmeta - 2:nmeta]
    varnames = [vn.strip() for vn in spaces.split(varline)]
    units = [u.strip()[1:-1].strip() for u in spaces.split(unitline)]
    print(units)
    data = np.fromstring(
        ', '.join(datalines[nmeta:]), sep=',').reshape(-1, len(varnames))
    outf = PseudoNetCDFFile()
    for pk, pv in meta.items():
        setattr(outf, pk, pv)

    outf.createDimension('time', data.shape[0])
    for varname, unit, vals in zip(varnames, units, data.T):
        var = outf.createVariable(varname.replace(' ', '_'), 'f', ('time',))
        var.units = unit
        var.standard_name = varname
        var[:] = np.ma.masked_values(vals[:], 99999)

    return outf


if __name__ == '__main__':
    f = skysonde1sec('/Users/barronh/Downloads/skysonde1sec.txt')



from __future__ import print_function
__all__ = ['NO', 'NP', 'NOP', 'MO', 'MP', 'MdnO', 'MdnP', 'STDO',
           'STDP', 'RM', 'RMdn', 'MB', 'MdnB', 'WDMB', 'WDMdnB',
           'FB', 'MNB', 'MdnNB', 'NMB', 'NMdnB', 'USUTPB', 'PSUTMNPB',
           'PSUTMdnNPB', 'PSUTNMPB', 'PSUTNMdnPB', 'ME', 'MdnE', 'WDME',
           'WDMdnE', 'FE', 'MNE', 'MdnNE', 'NME', 'NMdnE', 'USUTPE',
           'PSUTMNPE', 'PSUTMdnNPE', 'PSUTNMPE', 'PSUTNMdnPE', 'R2',
           'RMSE', 'RMSEs', 'RMSEu', 'E1', 'IOA', 'd1', 'AC', 'WDIOA',
           'WDRMSE', 'WDAC']

from .pncload import createconsole
import numpy as np


def STDO(obs, mod, axis=None):
    """ Standard deviation of Observations """
    return np.ma.std(obs, axis=axis)


def STDP(obs, mod, axis=None):
    """ Standard deviation of Predictions """
    return np.ma.std(mod, axis=axis)


def MNB(obs, mod, axis=None):
    """ Mean Normalized Bias (%)"""
    return np.ma.masked_invalid((mod - obs) / obs).mean(axis=axis) * 100.


def MNE(obs, mod, axis=None):
    """ Mean Normalized Gross Error (%)"""
    ne = np.ma.abs(mod - obs) / obs
    return np.ma.masked_invalid(ne).mean(axis=axis) * 100.


def MdnNB(obs, mod, axis=None):
    """ Median Normalized Bias (%)"""
    nb = (mod - obs) / obs
    return np.ma.median(np.ma.masked_invalid(nb), axis=axis) * 100.


def MdnNE(obs, mod, axis=None):
    """ Median Normalized Gross Error (%)"""
    ne = np.ma.abs(mod - obs) / obs
    return np.ma.median(np.ma.masked_invalid(ne), axis=axis) * 100.


def NO(obs, mod, axis=None):
    """ N Observations (#)"""
    return (~np.ma.getmaskarray(obs)).sum(axis=axis)


def NOP(obs, mod, axis=None):
    """ N Observations/Prediction Pairs (#)"""
    obsc, modc = matchmasks(obs, mod)
    return (~np.ma.getmaskarray(obsc)).sum(axis=axis)


def NP(obs, mod, axis=None):
    """ N Predictions (#)"""
    return (~np.ma.getmaskarray(mod)).sum(axis=axis)


def MO(obs, mod, axis=None):
    """ Mean Observations (obs unit)"""
    return obs.mean(axis=axis)


def MP(obs, mod, axis=None):
    """ Mean Predictions (model unit)"""
    return mod.mean(axis=axis)


def MdnO(obs, mod, axis=None):
    """ Median Observations (obs unit)"""
    return np.ma.median(obs, axis=axis)


def MdnP(obs, mod, axis=None):
    """ Median Predictions (model unit)"""
    return np.ma.median(mod, axis=axis)


def RM(obs, mod, axis=None):
    """ Mean Ratio Observations/Predictions (none)"""
    return np.ma.masked_invalid(obs / mod).mean(axis=axis)


def RMdn(obs, mod, axis=None):
    """ Median Ratio Observations/Predictions (none)"""
    return np.ma.median(np.ma.masked_invalid(obs / mod), axis=axis)


def MB(obs, mod, axis=None):
    """ Mean Bias"""
    return (mod - obs).mean(axis=axis)


def MdnB(obs, mod, axis=None):
    """ Median Bias"""
    return np.ma.median(mod - obs, axis=axis)


def WDMB(obs, mod, axis=None):
    """ Wind Direction Mean Bias"""
    return circlebias(mod - obs).mean(axis=axis)


def WDMdnB(obs, mod, axis=None):
    """ Wind Direction Median Bias"""
    return np.ma.median(circlebias(mod - obs), axis=axis)


def NMB(obs, mod, axis=None):
    """ Normalized Mean Bias (%)"""
    return (mod - obs).sum(axis=axis) / obs.sum(axis=axis) * 100.


def NMdnB(obs, mod, axis=None):
    """ Normalized Median Bias (%)"""
    mdnb = np.ma.median(mod - obs, axis=axis)
    mdno = np.ma.median(obs, axis=axis)
    return mdnb / mdno * 100.


def FB(obs, mod, axis=None):
    """ Fractional Bias (%)"""
    halffb = (mod - obs) / (mod + obs)
    return ((np.ma.masked_invalid(halffb)).mean(axis=axis) * 2.) * 100.


def ME(obs, mod, axis=None):
    """ Mean Gross Error (model and obs unit)"""
    return np.ma.abs(mod - obs).mean(axis=axis)


def MdnE(obs, mod, axis=None):
    """ Median Gross Error (model and obs unit)"""
    return np.ma.median(np.ma.abs(mod - obs), axis=axis)


def WDME(obs, mod, axis=None):
    """ Wind Direction Mean Gross Error (model and obs unit)"""
    return np.ma.abs(circlebias(mod - obs)).mean(axis=axis)


def WDMdnE(obs, mod, axis=None):
    """ Wind Direction Median Gross Error (model and obs unit)"""
    cb = circlebias(mod - obs)
    return np.ma.median(np.ma.abs(cb), axis=axis)


def NME(obs, mod, axis=None):
    """ Normalized Mean Error (%)"""
    out = (np.ma.abs(mod - obs).sum(axis=axis) / obs.sum(axis=axis)) * 100
    return out


def NMdnE(obs, mod, axis=None):
    """ Normalized Median Error (%)"""
    out = np.ma.median(np.ma.abs(mod - obs), axis=axis) / \
        np.ma.median(obs, axis=axis) * 100
    return out


def FE(obs, mod, axis=None):
    """ Fractional Error (%)"""
    return (np.ma.abs(mod - obs) / (mod + obs)).mean(axis=axis) * 2. * 100.


def USUTPB(obs, mod, axis=None):
    """ Unpaired Space/Unpaired Time Peak Bias (%)"""
    modmax = mod.max(axis=axis)
    obsmax = obs.max(axis=axis)
    return ((modmax - obsmax) / obsmax) * 100.


def USUTPE(obs, mod, axis=None):
    """ Unpaired Space/Unpaired Time Peak Error (%)"""
    obsmax = obs.max(axis=axis)
    return (np.ma.abs(mod.max(axis=axis) - obsmax) / obsmax) * 100.


def MNPB(obs, mod, paxis, axis=None):
    """ Mean Normalized Peak Bias (%)"""
    obsmax = obs.max(axis=paxis)
    return ((mod.max(axis=paxis) - obsmax) / obsmax).mean(axis=axis) * 100.


def MdnNPB(obs, mod, paxis, axis=None):
    """ Median Normalized Peak Bias (%)"""
    obsmax = obs.max(axis=paxis)
    modmax = mod.max(axis=paxis)
    return np.ma.median((modmax - obsmax) / obsmax, axis=axis) * 100.


def MNPE(obs, mod, paxis, axis=None):
    """ Mean Normalized Peak Error (%)"""
    modmax = mod.max(axis=paxis)
    obsmax = obs.max(axis=paxis)
    return ((np.ma.abs(modmax - obsmax)) / obsmax).mean(axis=axis) * 100.


def MdnNPE(obs, mod, paxis, axis=None):
    """ Median Normalized Peak Bias (%)"""
    modmax = mod.max(axis=paxis)
    obsmax = obs.max(axis=paxis)
    pe = np.ma.abs(modmax - obsmax)
    return np.ma.median((pe) / obsmax, axis=axis) * 100.


def NMPB(obs, mod, paxis, axis=None):
    """ Normalized Mean Peak Bias (%)"""
    modmax = mod.max(axis=paxis)
    obsmax = obs.max(axis=paxis)
    return (modmax - obsmax).mean(axis=axis) / obsmax.mean(axis=axis) * 100.


def NMdnPB(obs, mod, paxis, axis=None):
    """ Normalized Median Peak Bias (%)"""
    modmax = mod.max(axis=paxis)
    obsmax = obs.max(axis=paxis)
    return (np.ma.median((modmax - obsmax), axis=axis) /
            np.ma.median(obsmax, axis=axis) * 100.)


def NMPE(obs, mod, paxis, axis=None):
    """ Normalized Mean Peak Error (%)"""
    modmax = mod.max(axis=paxis)
    obsmax = obs.max(axis=paxis)
    return (np.ma.abs(modmax - obsmax).mean(axis=axis) /
            obsmax.mean(axis=axis) * 100.)


def NMdnPE(obs, mod, paxis, axis=None):
    """ Normalized Median Peak Bias (%)"""
    modmax = mod.max(axis=paxis)
    obsmax = obs.max(axis=paxis)
    return (np.ma.median(np.ma.abs(modmax - obsmax), axis=axis) /
            np.ma.median(obsmax, axis=axis) * 100.)


def PSUTMNPB(obs, mod, axis=None):
    """ Paired Space/Unpaired Time Mean Normalized Peak Bias (%)"""
    return MNPB(obs, mod, paxis=0, axis=None)


def PSUTMdnNPB(obs, mod, axis=None):
    """ Paired Space/Unpaired Time Median Normalized Peak Bias (%)"""
    return MdnNPB(obs, mod, paxis=0, axis=None)


def PSUTMNPE(obs, mod, axis=None):
    """ Paired Space/Unpaired Time Mean Normalized Peak Error (%)"""
    return MNPE(obs, mod, paxis=0, axis=None)


def PSUTMdnNPE(obs, mod, axis=None):
    """ Paired Space/Unpaired Time Median Normalized Peak Error (%)"""
    return MdnNPE(obs, mod, paxis=0, axis=None)


def PSUTNMPB(obs, mod, axis=None):
    """ Paired Space/Unpaired Time Normalized Mean Peak Bias (%)"""
    return NMPB(obs, mod, paxis=0, axis=None)


def PSUTNMPE(obs, mod, axis=None):
    """ Paired Space/Unpaired Time Normalized Mean Peak Error (%)"""
    return NMPE(obs, mod, paxis=0, axis=None)


def PSUTNMdnPB(obs, mod, axis=None):
    """ Paired Space/Unpaired Time Normalized Median Peak Bias (%)"""
    return NMdnPB(obs, mod, paxis=0, axis=None)


def PSUTNMdnPE(obs, mod, axis=None):
    """ Paired Space/Unpaired Time Normalized Median Peak Error (%)"""
    return NMdnPE(obs, mod, paxis=0, axis=None)


def R2(obs, mod, axis=None):
    """ Coefficient of Determination (unit squared)"""
    from scipy.stats.mstats import pearsonr
    if axis is None:
        return pearsonr(obs, mod)[0]**2
    else:
        r = pearsonr
        return apply_along_axis_2v(lambda x, y: r(x, y)[0]**2, axis, obs, mod)


def apply_along_axis_2v(func, axis, v1, v2, *args):
    v1r = np.rollaxis(v1, axis, v1.ndim)
    v2r = np.rollaxis(v2, axis, v2.ndim)
    result = np.array([func(v1r[idx], v2r[idx])
                       for idx in np.ndindex(v2r.shape[:1])])
    result = np.rollaxis(result[..., None], result.ndim, axis)
    return result


def RMSE(obs, mod, axis=None):
    """ Root Mean Square Error (model unit)"""
    return np.ma.sqrt(((mod - obs)**2).mean(axis=axis))


def WDRMSE(obs, mod, axis=None):
    """ Wind Direction Root Mean Square Error (model unit)"""
    return np.ma.sqrt(((circlebias(mod - obs))**2).mean(axis=axis))


def RMSEs(obs, mod, axis=None):
    """Root Mean Squared Error systematic (obs, mod_hat)"""
    from scipy.stats.mstats import linregress
    if axis is None:
        try:
            m, b, rval, pval, stderr = linregress(obs, mod)
            mod_hat = b + m * obs
            return RMSE(obs, mod_hat)
        except ValueError:
            return None
    else:
        myvals = apply_along_axis_2v(
            lambda x, y: linregress(x, y), axis, obs, mod)
        myvals = np.rollaxis(myvals, myvals.ndim - 1, 0).astype(obs.dtype)
        m, b, rval, pval, stderr = myvals
        mod_hat = b + m * obs
        result = RMSE(obs, mod_hat, axis=axis)
        return result


def matchmasks(a1, a2):
    mask = np.ma.getmaskarray(a1) | np.ma.getmaskarray(a2)
    return np.ma.masked_where(mask, a1), np.ma.masked_where(mask, a2)


def matchedcompressed(a1, a2):
    a1, a2 = matchmasks(a1, a2)
    return a1.compressed(), a2.compressed()


def RMSEu(obs, mod, axis=None):
    """Root Mean Squared Error unsystematic (mod_hat, mod)"""
    from scipy.stats.mstats import linregress
    if axis is None:
        try:
            m, b, rval, pval, stderr = linregress(obs, mod)
            mod_hat = b + m * obs
            return RMSE(mod_hat, mod)
        except ValueError:
            return None
    else:
        myvals = apply_along_axis_2v(
            lambda x, y: linregress(x, y), axis, obs, mod)
        myvals = np.rollaxis(myvals, myvals.ndim - 1, 0).astype(obs.dtype)
        m, b, rval, pval, stderr = myvals
        mod_hat = b + m * obs
        result = RMSE(mod_hat, mod, axis=axis)
        return result


def d1(obs, mod, axis=None):
    """ Modified Index of Agreement, d1"""
    return (1.0 -
            np.ma.abs(obs - mod).sum(axis=axis) /
            (np.ma.abs(mod - obs.mean(axis=axis)) +
             np.ma.abs(obs - obs.mean(axis=axis))).sum(axis=axis))


def E1(obs, mod, axis=None):
    """ Modified Coefficient of Efficiency, E1"""
    return (1.0 -
            np.ma.abs(obs - mod).sum(axis=axis) /
            (np.ma.abs(obs - obs.mean(axis=axis))).sum(axis=axis))


def IOA(obs, mod, axis=None):
    """ Index of Agreement, IOA"""
    obsmean = obs.mean(axis=axis)
    if axis is not None:
        obsmean = np.expand_dims(obsmean, axis=axis)
    return (1.0 -
            (np.ma.abs(obs - mod)**2).sum(axis=axis) /
            ((np.ma.abs(mod - obsmean) +
              np.ma.abs(obs - obsmean))**2).sum(axis=axis))


def circlebias(b):
    b = np.ma.where(b > 180, b - 360, b)
    b = np.ma.where(b < -180, b + 360, b)
    return b


def WDIOA(obs, mod, axis=None):
    """ Wind Direction Index of Agreement, IOA"""
    obsmean = obs.mean(axis=axis)
    if axis is not None:
        obsmean = np.expand_dims(obsmean, axis=axis)
    b = circlebias(mod - obs)

    bhat = circlebias(mod - obsmean)

    ohat = circlebias(obs - obsmean)

    return (1.0 - (np.ma.abs(b)**2).sum(axis=axis) /
            ((np.ma.abs(bhat) + np.ma.abs(ohat))**2).sum(axis=axis))


def AC(obs, mod, axis=None):
    """ Anomaly Correlation """
    obs_bar = obs.mean(axis=axis)
    if axis is not None:
        obs_bar = np.expand_dims(obs_bar, axis=axis)
    p1 = ((mod - obs_bar) * (obs - obs_bar)).sum(axis=axis)
    p2 = (((mod - obs_bar)**2).sum(axis=axis) *
          ((obs - obs_bar)**2).sum(axis=axis))**0.5
    return p1 / p2


def WDAC(obs, mod, axis=None):
    """ Wind Direction Anomaly Correlation """
    obs_bar = obs.mean(axis=axis)
    if axis is not None:
        obs_bar = np.expand_dims(obs_bar, axis=axis)
    p1 = (circlebias(mod - obs_bar) *
          circlebias(obs - obs_bar)).sum(axis=axis)
    p2 = ((circlebias(mod - obs_bar)**2).sum(axis=axis) *
          (circlebias(obs - obs_bar)**2).sum(axis=axis))**0.5
    return p1 / p2


def stat_spatial(ifile0, ifile1, funcs=__all__,
                 variables=['O3'], counties=False):
    """
    ifile0 - left hand side of equation
    ifile1 - right hand side of equation
    variables - list of variables to plot
    """
    from mpl_toolkits.basemap import Basemap
    from matplotlib.pyplot import figure, show
    lonlatcoords = getattr(ifile0, 'lonlatcoords',
                           getattr(ifile1, 'lonlatcoords', ''))
    lon, lat = np.array(
        map(lambda x: map(float, x.split(',')), lonlatcoords.split('/'))).T
    latmin, latmax = lat.min(), lat.max()
    lonmin, lonmax = lon.min(), lon.max()
    bmap = Basemap(llcrnrlat=latmin, llcrnrlon=lonmin,
                   urcrnrlat=latmax, urcrnrlon=lonmax, resolution='i')
    for vark in variables:
        for statname in funcs:
            statfunc = eval(statname)
            fig = figure()
            ax = fig.add_subplot(111)
            ax.set_title(vark + ' ' + statname)
            var_0 = ifile0.variables[vark]
            var_1 = ifile1.variables[vark]
            try:
                pidx = list(var_0.dimensions).index('points')
            except Exception:
                pidx = list(var_1.dimensions).index('points')
            statvs = []
            for sitei in range(var_0.shape[pidx]):
                val_0 = var_0[:].take([sitei], axis=pidx).ravel()
                val_1 = var_1[:].take([sitei], axis=pidx).ravel()
                statvs.append(statfunc(val_0, val_1))
            dots = bmap.scatter(x=lon, y=lat, c=statvs, ax=ax, cmap='jet')

            bmap.drawcoastlines(ax=ax)
            bmap.drawcountries(ax=ax)
            bmap.drawstates(ax=ax)
            fig.colorbar(dots)
            if counties:
                bmap.counties(ax=ax)
            show()


def stat_timeseries(ifile0, ifile1, variables=['O3'], counties=False):
    from pylab import figure, show
    for vark in variables:
        for statname in __all__:
            statfunc = eval(statname)
            fig = figure()
            ax = fig.add_subplot(111)
            ax.set_title(vark + ' ' + statname)
            var_0 = ifile0.variables[vark]
            var_1 = ifile1.variables[vark]
            tidx = list(var_0.dimensions).index('time')
            statvs = []
            for timei in range(var_0.shape[tidx]):
                val_0 = var_0[:].take([timei], axis=tidx).ravel()
                val_1 = var_1[:].take([timei], axis=tidx).ravel()
                statvs.append(statfunc(val_0, val_1))
            ax.plot(statvs)
            show()


def pnceval(args):
    from warnings import warn
    from collections import OrderedDict
    from PseudoNetCDF.core._functions import pncbfunc
    if args.variables is None:
        args.variables = set(
            args.ifiles[0].variables.keys()).difference(args.coordkeys)
    console = createconsole(args.ifiles, args)
    ifile0, ifile1 = args.ifiles
    from PseudoNetCDF.coordutil import gettimes
    print('# ifile0=' + args.ipath[0])
    print('# ifile1=' + args.ipath[1])
    print('# Stats calculated as func(ifile0, ifile1)')
    print('# Generally: func(obs, mod, ....)')
    for ipath, ifile in zip(args.ipath, args.ifiles):
        print('# Meta-data from %s' % ipath)
        try:
            times = gettimes(ifile)
            tstart = times[:].min()
            tstop = times[:].max()
            print('# Date Range: ' + str(tstart) + ' to ' + str(tstop))
        except Exception as e:
            warn(str(e))
        try:
            lon = ifile.variables['longitude']
            print('# Longitude Range: ' + str(float(lon.min())) +
                  ' to ' + str(float(lon.max())))
        except Exception as e:
            warn(str(e))
        try:
            lat = ifile.variables['latitude']
            print('# Latitude Range: ' + str(float(lat.min())) +
                  ' to ' + str(float(lat.max())))
        except Exception as e:
            warn(str(e))

    np.seterr(divide='ignore', invalid='ignore')
    output = OrderedDict()
    for k in args.funcs:
        console.locals[k] = func = eval(k)
        output[k] = OrderedDict()
        if args.csv:
            print('# %s: %s' % (k, func.__doc__.strip()))
        try:
            console.locals[k + '_f'] = ofile = pncbfunc(func, ifile0, ifile1)
        except Exception as e:
            warn("Skipped " + k + ';' + str(e))
            continue
        for vk in args.variables:
            if vk in ('time', 'TFLAG'):
                continue
            if args.csv:
                outv = ofile.variables[vk][:]
                outv = np.ma.array(outv)
                output[k][vk] = outv.ravel()[0]
            else:
                print('%s,%s,%s,%f' % (vk, func.__doc__.strip(),
                                       k, ofile.variables[vk].ravel()[0]))
    if args.csv:
        print(','.join(['VAR'] + args.funcs))
        for vk in args.variables:
            print(','.join([vk] + ['%f' % output[fk].get(vk, np.nan)
                                   for fk in args.funcs]))

    np.seterr(divide='warn', invalid='warn')
    if args.interactive:
        console.interact()


def main():
    from .pncparse import getparser, pncparse

    parser = getparser(has_ofile=False, plot_options=False, interactive=True)
    parser.add_argument('--csv', dest='csv', default=False,
                        action='store_true',
                        help='Print all data in CSV format')
    parser.add_argument('--funcs', default=__all__,
                        type=lambda x: x.split(','),
                        help=('Functions to evaluate split by , ' +
                              '(default: %s)') % ','.join(__all__))

    ifiles, args = pncparse(has_ofile=False, interactive=True, parser=parser)
    args.ifiles = ifiles
    pnceval(args)


if __name__ == '__main__':
    main()



__doc__ = """
.. _racmfiles
:mod:`racmfile` -- RACM File Interfaces
===================================

.. module:: RACM
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF`  random access read
              based file interfaces for all RACM related files.  See
              PseudoNetCDF.sci_var.PseudoNetCDFFile for interface details.
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""
__all__ = ['boxmodel']

from . import boxmodel



from __future__ import print_function
try:
    from setuptools import setup
except Exception:
    from distutils.core import setup
import os
import subprocess

join = os.path.join


def find_packages():
    import os
    packages = []
    walker = os.walk('src')
    prefix = join(os.path.curdir, 'src')
    for thisdir, itsdirs, itsfiles in walker:
        if '__init__.py' in itsfiles:
            packages.append(thisdir[len(prefix) - 1:])

    return packages


def find_data():
    import os
    import re
    suffixes = ['yaml', 'nc', 'net', 'irr', 'phy', 'ptb',
                'sum', 'voc', 'txt', 'xls', 'graffle']
    data_pattern = re.compile(
        r'.*(.|_)(' + '|'.join(suffixes) + ')$')
    data = []
    prefix = join(os.path.curdir, 'src', 'PseudoNetCDF')
    walker = os.walk('src')
    for thisdir, itsdirs, itsfiles in walker:
        if thisdir != os.path.join('src',
                                   'PseudoNetCDF.egg-info'):
            data.extend([join(thisdir[len(prefix) - 1:], f)
                         for f in itsfiles
                         if data_pattern.match(f) is not None])

    return data


packages = find_packages()
data = find_data()
MAJOR = 3
MINOR = 2
MICRO = 0
ISRELEASED = False
VERSION = '%d.%d.%d' % (MAJOR, MINOR, MICRO)


# Return the git revision as a string
def git_version():
    # Copied from scipy
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ['SYSTEMROOT', 'PATH']:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env['LANGUAGE'] = 'C'
        env['LANG'] = 'C'
        env['LC_ALL'] = 'C'
        Popen = subprocess.Popen
        out = Popen(cmd, stdout=subprocess.PIPE, env=env).communicate()[0]
        return out

    try:
        out = _minimal_ext_cmd(['git', 'rev-parse', 'HEAD'])
        GIT_REVISION = out.strip().decode('ascii')
    except OSError:
        GIT_REVISION = "Unknown"

    return GIT_REVISION


def get_version_info():
    # Modified from scipy
    # Adding the git rev number needs to be done inside
    # write_version_py(), otherwise the import of
    # PseudoNetCDF.version messes up the build under Python 3.
    FULLVERSION = VERSION
    if os.path.exists('.git'):
        GIT_REVISION = git_version()
    elif os.path.exists('src/PseudoNetCDF/version.py'):
        # must be a source distribution, use existing version file
        # load as separate module to not load PseudoNetCDF/__init__.py
        import imp
        version = imp.load_source('PseudoNetCDF.version',
                                  'src/PseudoNetCDF/version.py')
        GIT_REVISION = version.git_revision
    else:
        GIT_REVISION = "Unknown"

    if not ISRELEASED:
        FULLVERSION += '.dev0+' + GIT_REVISION[:7]

    return FULLVERSION, GIT_REVISION


def write_version_py(filename='src/PseudoNetCDF/version.py'):
    # Modified from scipy
    cnt = """
# THIS FILE IS GENERATED FROM PseudoNetCDF SETUP.PY
short_version = '%(version)s'
version = '%(version)s'
full_version = '%(full_version)s'
git_revision = '%(git_revision)s'
release = %(isrelease)s
if not release:
    version = full_version
"""
    FULLVERSION, GIT_REVISION = get_version_info()

    a = open(filename, 'w')
    try:
        a.write(cnt % {'version': VERSION,
                       'full_version': FULLVERSION,
                       'git_revision': GIT_REVISION,
                       'isrelease': str(ISRELEASED)})
    finally:
        a.close()


write_version_py(filename='src/PseudoNetCDF/version.py')

long_desc = """NetCDF, NCO, and CDO are fantastic softwares that I use,
emulate, and admire. The primary drawbacks are restrictions on which
scientific data sources they will and won't work with, and what types
of operations they will and won't do. PseudoNetCDF was originally just
a NetCDF-like interface for many data formats, but has grown to include
many functionalities from NCO and CDO. This is a platform independent,
easy to install software to make scientific data manipulation easy."""

script_list = ['scripts/pncmadis2pnceval.py', 'scripts/pncaqsraw4pnceval.py',
               'scripts/pncaqsrest4pnceval.py', 'scripts/pncasos4pnceval.py',
               'scripts/pnc1d.py', 'scripts/pnc2d.py',
               'scripts/pncboundaries.py', 'scripts/pncdiurnal.py',
               'scripts/pncdump', 'scripts/pncdump.py', 'scripts/pnceval',
               'scripts/pnceval.py', 'scripts/pncgen', 'scripts/pncgen.py',
               'scripts/pncglobal2cmaq.py', 'scripts/pncload',
               'scripts/pncmap.py', 'scripts/pncqq.py',
               'scripts/pncscatter.py', 'scripts/pncts.py',
               'scripts/pncvertprofile.py', 'scripts/pncview',
               'scripts/pncview.py', 'scripts/pncwindrose.py']

requires_list = ['numpy>=1.2', 'netCDF4', 'scipy', 'matplotlib', 'pyyaml']
extra_requires_dict = {'textfiles': ['pandas'],
                       'projections': ['pyproj'],
                       'mapping': ['basemap'],
                       }

setup(name='PseudoNetCDF',
      version=VERSION,
      author='Barron Henderson',
      author_email='barronh@gmail.com',
      maintainer='Barron Henderson',
      maintainer_email='barronh@gmail.com',
      description=('Like NetCDF and NCO, but works with NetCDF and other ' +
                   'scientific formats.'),
      long_description=long_desc,
      packages=packages,
      package_dir={'': 'src'},
      package_data={'PseudoNetCDF': data},
      scripts=script_list,
      install_requires=requires_list,
      extras_require=extra_requires_dict,
      url='http://github.com/barronh/pseudonetcdf/',
      classifiers=['Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3',
                   'Operating System :: MacOS',
                   'Operating System :: Microsoft :: Windows',
                   'Operating System :: POSIX',
                   'Topic :: Scientific/Engineering',
                   'Topic :: Scientific/Engineering :: Atmospheric Science',
                   ]
      )
